import os
import sys
import time
from PyQt5.QtWidgets import (
    QDialog, QDialogButtonBox, QTreeWidgetItem,
    QCompleter, QMessageBox, QFileDialog, QWidget)
from PyQt5.QtGui import QIcon, QPixmap
from PyQt5.QtCore import Qt, QTimer

import ray
from gui_server_thread import GUIServerThread
from gui_tools import (default_session_root, ErrDaemon, _translate,
                       CommandLineArgs, RS, isDarkTheme)

import ui_open_session
import ui_new_session
import ui_list_snapshots
import ui_save_template_session
import ui_nsm_open_info
import ui_abort_session
import ui_about_raysession
import ui_add_application
import ui_donations
import ui_new_executable
import ui_error_dialog
import ui_quit_app
import ui_client_properties
import ui_stop_client
import ui_stop_client_no_save
import ui_abort_copy
import ui_client_trash
import ui_daemon_url
import ui_edit_executable
import ui_snapshot_progress
import ui_waiting_close_user

class ChildDialog(QDialog):
    def __init__(self, parent):
        QDialog.__init__(self, parent)
        self._session = parent._session
        self._signaler = self._session._signaler

        daemon_manager = self._session._daemon_manager
        self.daemon_launched_before = daemon_manager.launched_before

        self._signaler.server_status_changed.connect(self.serverStatusChanged)
        self._signaler.server_copying.connect(self.serverCopying)

        self.server_copying = parent.server_copying

    def toDaemon(self, *args):
        server = GUIServerThread.instance()
        if server:
            server.toDaemon(*args)
        else:
            sys.stderr.write('Error No GUI OSC Server, can not send %s.\n'
                             % args)

    def serverStatusChanged(self, server_status):
        return

    def serverCopying(self, bool_copying):
        self.server_copying = bool_copying
        self.serverStatusChanged(self._session.server_status)
        
    def changeRootFolder(self):
        root_folder = QFileDialog.getExistingDirectory(
            self, 
            _translate("root_folder_dialogs",
                       "Choose root folder for sessions"), 
            CommandLineArgs.session_root, 
            QFileDialog.ShowDirsOnly)
        
        if not root_folder:
            return
        
        # Security, kde dialogs sends $HOME if user type a folder path
        # that doesn't already exists.
        if os.getenv('HOME') and root_folder == os.getenv('HOME'):
            return
        
        errorDialog = QMessageBox(
            QMessageBox.Critical, 
            _translate('root_folder_dialogs', 'unwritable dir'), 
            _translate(
                'root_folder_dialogs',
                '<p>You have no permissions for %s,<br>' % root_folder \
                    + 'choose another directory !</p>'))
        
        if not os.path.exists(root_folder):
            try:
                os.makedirs(root_folder)
            except:
                errorDialog.exec()
                return
        
        if not os.access(root_folder, os.W_OK):
            errorDialog.exec()
            return
        
        RS.settings.setValue('default_session_root', root_folder)
        self.toDaemon('/ray/server/change_root', root_folder)

    def leaveEvent(self, event):
        if self.isActiveWindow():
            self.parent().mouse_is_inside = False
        QDialog.leaveEvent(self, event)

    def enterEvent(self, event):
        self.parent().mouse_is_inside = True
        QDialog.enterEvent(self, event)


class SessionItem(QTreeWidgetItem):
    def __init__(self, list):
        QTreeWidgetItem.__init__(self, list)
        
    def showConditionnaly(self, string):
        show = bool(string.lower() in self.data(0, Qt.UserRole).lower())
        
        n=0
        for i in range(self.childCount()):
            if self.child(i).showConditionnaly(string.lower()):
                n+=1
        if n:
            show = True
            
        self.setExpanded(bool(n and string))
        self.setHidden(not show)
        return show
    
    def findItemWith(self, string):
        if self.data(0, Qt.UserRole) == string:
            return self
        
        item = None
        
        for i in range(self.childCount()):
            item = self.child(i).findItemWith(string)
            if item:
                break
            
        return item
        
    def __lt__(self, other):
        if self.childCount() and not other.childCount():
            return True
        
        if other.childCount() and not self.childCount():
            return False
        
        return bool(self.text(0).lower() < other.text(0).lower())

class SessionFolder:
    name = ""
    has_session = True
    path = ""
    
    def __init__(self, name):
        self.name = name
        self.subfolders = []
    
    def setPath(self, path):
        self.path = path
    
    def makeItem(self):
        item = SessionItem([self.name])
        
        item.setData(0, Qt.UserRole, self.path)
        if self.subfolders:
            item.setIcon(0, QIcon.fromTheme('folder'))
        
        if not self.path:
            item.setFlags(item.flags() & ~Qt.ItemIsSelectable)
        
        for folder in self.subfolders:
            sub_item = folder.makeItem()
            item.addChild(sub_item)
            
        return item
    
class OpenSessionDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_open_session.Ui_DialogOpenSession()
        self.ui.setupUi(self)

        self.ui.toolButtonFolder.clicked.connect(self.changeRootFolder)
        self.ui.sessionList.currentItemChanged.connect(
            self.currentItemChanged)
        self.ui.sessionList.setFocus(Qt.OtherFocusReason)
        self.ui.sessionList.itemDoubleClicked.connect(self.goIfAny)
        self.ui.sessionList.itemClicked.connect(self.deployItem)
        self.ui.filterBar.textEdited.connect(self.updateFilteredList)
        self.ui.filterBar.updownpressed.connect(self.updownPressed)
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.ui.currentSessionsFolder.setText(CommandLineArgs.session_root)
        
        self._signaler.add_sessions_to_list.connect(self.addSessions)
        self._signaler.root_changed.connect(self.rootChanged)

        self.toDaemon('/ray/server/list_sessions', 0)

        if self.daemon_launched_before:
            self.ui.toolButtonFolder.setVisible(False)
            self.ui.currentSessionsFolder.setVisible(False)
            self.ui.labelSessionsFolder.setVisible(False)

        self.server_will_accept = False
        self.has_selection = False

        self.serverStatusChanged(self._session.server_status)
        
        self.folders = []
        self.all_items = []
        
        self._last_mouse_click = 0
        

    def serverStatusChanged(self, server_status):
        self.ui.toolButtonFolder.setEnabled(
            bool(server_status in (ray.ServerStatus.OFF,
                                   ray.ServerStatus.READY,
                                   ray.ServerStatus.CLOSE)))

        self.server_will_accept = bool(
            server_status in (
                ray.ServerStatus.OFF,
                ray.ServerStatus.READY) and not self.server_copying)
        self.preventOk()
        
    def rootChanged(self, session_root):
        self.ui.currentSessionsFolder.setText(session_root)
        self.ui.sessionList.clear()
        self.folders.clear()
        self.toDaemon('/ray/server/list_sessions', 0)

    def addSessions(self, session_names):
        for session_name in session_names:
            folder_div = session_name.split('/')
            folders = self.folders
            
            for i in range(len(folder_div)):
                f = folder_div[i]
                for g in folders:
                    if g.name == f:
                        if i+1 == len(folder_div):
                            g.setPath(session_name)
                            
                        folders = g.subfolders
                        break
                else:
                    new_folder = SessionFolder(f)
                    if i+1 == len(folder_div):
                        new_folder.setPath(session_name)
                    folders.append(new_folder)
                    folders = new_folder.subfolders
                        
        self.ui.sessionList.clear()
        
        for folder in self.folders:
            item = folder.makeItem()
            self.ui.sessionList.addTopLevelItem(item)
        
        self.ui.sessionList.sortByColumn(0, Qt.AscendingOrder)
        
        # Try to select last used session
        root_item = self.ui.sessionList.invisibleRootItem()
        for i in range(root_item.childCount()):
            item = root_item.child(i)
            last_session_item = item.findItemWith(
                RS.settings.value('last_session', type=str))
            
            if last_session_item:
                self.ui.sessionList.setCurrentItem(last_session_item)
                self.ui.sessionList.scrollToItem(last_session_item)
                break

    def updateFilteredList(self, filt):
        filter_text = self.ui.filterBar.displayText()
        root_item = self.ui.sessionList.invisibleRootItem()

        ## hide all non matching items
        for i in range(root_item.childCount()):
            root_item.child(i).showConditionnaly(filter_text)

        # if selected item not in list, then select the first visible
        if (not self.ui.sessionList.currentItem()
                or self.ui.sessionList.currentItem().isHidden()):
            for i in range(root_item.childCount()):
                item = root_item.child(i)
                if not item.isHidden():
                    self.ui.sessionList.setCurrentItem(item)
                    break

        if (not self.ui.sessionList.currentItem()
                or self.ui.sessionList.currentItem().isHidden()):
            self.ui.filterBar.setStyleSheet(
                "QLineEdit { background-color: red}")
            self.ui.sessionList.setCurrentItem(None)
        else:
            self.ui.filterBar.setStyleSheet("")
            self.ui.sessionList.scrollTo(self.ui.sessionList.currentIndex())

    def updownPressed(self, key):
        root_item = self.ui.sessionList.invisibleRootItem()
        row = self.ui.sessionList.currentIndex().row()
        
        if key == Qt.Key_Up:
            if row == 0:
                return
            row -= 1
            while root_item.child(row).isHidden():
                if row == 0:
                    return
                row -= 1
        elif key == Qt.Key_Down:
            if row == root_item.childCount() - 1:
                return
            row += 1
            while root_item.child(row).isHidden():
                if row == root_item.childCount() - 1:
                    return
                row += 1
                
        self.ui.sessionList.setCurrentItem(root_item.child(row))

    def currentItemChanged(self, item, previous_item):
        self.has_selection = bool(item and item.data(0, Qt.UserRole))
        self.preventOk()

    def preventOk(self):
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(
            bool(self.server_will_accept and self.has_selection))

    def getSelectedSession(self):
        if self.ui.sessionList.currentItem():
            return self.ui.sessionList.currentItem().data(0, Qt.UserRole)
    
    def deployItem(self, item, column):
        if not item.childCount():
            return 
        
        if time.time() - self._last_mouse_click > 0.35:
            item.setExpanded(not item.isExpanded())
            
        self._last_mouse_click = time.time()
        
    def goIfAny(self, item, column):
        if item.childCount():
            return 
        
        if (self.server_will_accept and self.has_selection
            and self.ui.sessionList.currentItem().data(0, Qt.UserRole)):
                self.accept()
            

class NewSessionDialog(ChildDialog):
    def __init__(self, parent, duplicate_window=False):
        ChildDialog.__init__(self, parent)
        self.ui = ui_new_session.Ui_DialogNewSession()
        self.ui.setupUi(self)

        self.is_duplicate = bool(duplicate_window)

        self.ui.currentSessionsFolder.setText(CommandLineArgs.session_root)
        self.ui.toolButtonFolder.clicked.connect(self.changeRootFolder)
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.ui.lineEdit.setFocus(Qt.OtherFocusReason)
        self.ui.lineEdit.textChanged.connect(self.textChanged)

        self.session_list = []
        self.template_list = []
        self.sub_folders = []

        self._signaler.server_status_changed.connect(self.serverStatusChanged)
        self._signaler.add_sessions_to_list.connect(self.addSessionsToList)
        self._signaler.session_template_found.connect(self.addTemplatesToList)
        self._signaler.root_changed.connect(self.rootChanged)
        
        self.toDaemon('/ray/server/list_sessions', 1)
        
        if self.is_duplicate:
            self.ui.labelTemplate.setVisible(False)
            self.ui.comboBoxTemplate.setVisible(False)
            self.ui.labelNewSessionName.setText(
                _translate('Duplicate', 'Duplicated session name :'))
            self.setWindowTitle(_translate('Duplicate', 'Duplicate Session'))
        else:
            self.toDaemon('/ray/server/list_session_templates')

        if self.daemon_launched_before:
            self.ui.toolButtonFolder.setVisible(False)
            self.ui.currentSessionsFolder.setVisible(False)
            self.ui.labelSessionsFolder.setVisible(False)

        self.initTemplatesComboBox()
        self.setLastTemplateSelected()
        
        self.ui.labelSubFolder.setVisible(False)
        self.ui.comboBoxSubFolder.setVisible(False)

        self.server_will_accept = False
        self.text_is_valid = False

        self.serverStatusChanged(self._session.server_status)

    def serverStatusChanged(self, server_status):
        self.ui.toolButtonFolder.setEnabled(
            bool(server_status == ray.ServerStatus.OFF))

        self.server_will_accept = bool(
            server_status in (
                ray.ServerStatus.OFF,
                ray.ServerStatus.READY) and not self.server_copying)
        if self.is_duplicate:
            self.server_will_accept = bool(
                server_status == ray.ServerStatus.READY and not self.server_copying)

        self.preventOk()
    
    def rootChanged(self, session_root):
        self.ui.currentSessionsFolder.setText(session_root)
        self.session_list.clear()
        self.sub_folders.clear()
        self.initSubFolderCombobox()
        self.toDaemon('/ray/server/list_sessions', 1)
    
    def initTemplatesComboBox(self):
        self.ui.comboBoxTemplate.clear()
        self.ui.comboBoxTemplate.addItem(
            _translate('session_template', "empty"))
        self.ui.comboBoxTemplate.addItem(
            _translate(
                'session_template',
                "with JACK patch memory"))
        self.ui.comboBoxTemplate.insertSeparator(2)

    def initSubFolderCombobox(self):
        self.ui.comboBoxSubFolder.clear()
        self.ui.comboBoxSubFolder.addItem(_translate('new_session', 'none'))
        
        for sub_folder in self.sub_folders:
            self.ui.comboBoxSubFolder.addItem(QIcon.fromTheme('folder'), 
                                              sub_folder)
        
        self.ui.labelSubFolder.setVisible(bool(self.sub_folders))
        self.ui.comboBoxSubFolder.setVisible(bool(self.sub_folders))
        
    def setLastTemplateSelected(self):
        last_used_template = RS.settings.value('last_used_template', type=str)

        if last_used_template.startswith('///'):
            if last_used_template == '///withJACKPATCH':
                self.ui.comboBoxTemplate.setCurrentIndex(1)
        else:
            if last_used_template in self.template_list:
                self.ui.comboBoxTemplate.setCurrentText(last_used_template)

        if not last_used_template:
            self.ui.comboBoxTemplate.setCurrentIndex(1)

    def setLastSubFolderSelected(self):
        last_subfolder = RS.settings.value('last_subfolder', type=str)
        
        if last_subfolder:
            self.ui.comboBoxSubFolder.setCurrentText(last_subfolder)
        else:
            self.ui.comboBoxSubFolder.setCurrentIndex(0)
    
    def addSessionsToList(self, session_names):
        self.session_list += session_names
        
        for session_name in session_names:
            if '/' in session_name:
                new_dir = os.path.dirname(session_name)
                if not new_dir in self.sub_folders:
                    self.sub_folders.append(new_dir)
        
        self.sub_folders.sort()
        self.initSubFolderCombobox()
        self.setLastSubFolderSelected()

    def addTemplatesToList(self, template_list):
        for template in template_list:
            if template not in self.template_list:
                self.template_list.append(template)

        if not self.template_list:
            return

        self.template_list.sort()

        self.initTemplatesComboBox()

        for template_name in self.template_list:
            self.ui.comboBoxTemplate.addItem(template_name)

        self.setLastTemplateSelected()

    def getSessionName(self):
        if self.ui.comboBoxSubFolder.currentIndex() > 0:
            return '%s/%s' % (self.ui.comboBoxSubFolder.currentText(),
                              self.ui.lineEdit.text())
        else:
            return self.ui.lineEdit.text()

    def getTemplateName(self):
        if self.ui.comboBoxTemplate.currentIndex() == 0:
            return ""

        if self.ui.comboBoxTemplate.currentIndex() == 1:
            return '///withJACKPATCH'

        return self.ui.comboBoxTemplate.currentText()
    
    def getSubFolder(self):
        if self.ui.comboBoxSubFolder.currentIndex() == 0:
            return ""
        
        return  self.ui.comboBoxSubFolder.currentText()
    
    def textChanged(self, text):
        full_session_text = text
        if self.ui.comboBoxSubFolder.currentIndex():
            full_session_text = "%s/%s" % (
                                    self.ui.comboBoxSubFolder.currentText(),
                                    text)
            
        self.text_is_valid = bool(text
                                  and full_session_text 
                                        not in self.session_list)
        self.preventOk()

    def preventOk(self):
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(
            bool(self.server_will_accept and self.text_is_valid))


class AbstractSaveTemplateDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_save_template_session.Ui_DialogSaveTemplateSession()
        self.ui.setupUi(self)

        self.server_will_accept = False

        self.ui.lineEdit.textEdited.connect(self.textEdited)
        self.ui.pushButtonAccept.clicked.connect(self.verifyAndAccept)
        self.ui.pushButtonAccept.setEnabled(False)

    def textEdited(self, text):
        if '/' in text:
            self.ui.lineEdit.setText(text.replace('/', '⁄'))
        self.allowOkButton()

    def getTemplateName(self):
        return self.ui.lineEdit.text()

    def allowOkButton(self, text=''):
        self.ui.pushButtonAccept.setEnabled(
            bool(self.server_will_accept and self.ui.lineEdit.text()))

    def verifyAndAccept(self):
        template_name = self.getTemplateName()
        if template_name in self.template_list:
            ret = QMessageBox.question(
                self,
                _translate(
                    'session template',
                    'Overwrite Template ?'),
                _translate(
                    'session_template',
                    'Template <strong>%s</strong> already exists.\nOverwrite it ?') %
                template_name)
            if ret == QMessageBox.No:
                return
        self.accept()


class SaveTemplateSessionDialog(AbstractSaveTemplateDialog):
    def __init__(self, parent):
        AbstractSaveTemplateDialog.__init__(self, parent)

        self.template_list = []

        self._signaler.session_template_found.connect(self.addTemplatesToList)
        self.toDaemon('/ray/server/list_session_templates')

        self.serverStatusChanged(self._session.server_status)

    def serverStatusChanged(self, server_status):
        self.server_will_accept = bool(server_status == ray.ServerStatus.READY)

        if server_status == ray.ServerStatus.OFF:
            self.reject()

        self.allowOkButton()

    def addTemplatesToList(self, template_list):
        self.template_list += template_list


class SaveTemplateClientDialog(AbstractSaveTemplateDialog):
    def __init__(self, parent):
        AbstractSaveTemplateDialog.__init__(self, parent)

        self.template_list = []
        self.ui.pushButtonAccept.setEnabled(False)

        self.ui.labelNewTemplateName.setText(
            _translate(
                'new client template',
                "New application template name :"))

        self._signaler.user_client_template_found.connect(
            self.addTemplatesToList)

        self.toDaemon('/ray/server/list_user_client_templates')

        self.serverStatusChanged(self._session.server_status)

    def serverStatusChanged(self, server_status):
        self.server_will_accept = bool(
            server_status not in (
                ray.ServerStatus.OFF,
                ray.ServerStatus.CLOSE) and not self.server_copying)

        if server_status == ray.ServerStatus.OFF:
            self.reject()

        self.allowOkButton()

    def addTemplatesToList(self, template_list):
        for template in template_list:
            self.template_list.append(template.split('/')[0])


class EditExecutableDialog(ChildDialog):
    def __init__(self, parent, client):
        ChildDialog.__init__(self, parent)
        self.ui = ui_edit_executable.Ui_Dialog()
        self.ui.setupUi(self)

        self.ui.lineEditExecutable.setText(client.executable_path)
        self.ui.lineEditArguments.setText(client.arguments)

    def getExecutable(self):
        return self.ui.lineEditExecutable.text()

    def getArguments(self):
        return self.ui.lineEditArguments.text()


class ClientPropertiesDialog(ChildDialog):
    def __init__(self, parent, client):
        ChildDialog.__init__(self, parent)
        self.ui = ui_client_properties.Ui_Dialog()
        self.ui.setupUi(self)

        self.client = client

        self.ui.lineEditIcon.textEdited.connect(self.changeIconwithText)
        self.ui.pushButtonSaveChanges.clicked.connect(self.saveChanges)
        self.ui.toolButtonEditExecutable.clicked.connect(self.editExecutable)

    def updateContents(self):
        self.ui.labelExecutable.setText(self.client.executable_path)
        self.ui.labelArguments.setText(self.client.arguments)
        self.ui.labelId.setText(self.client.client_id)
        self.ui.labelClientName.setText(self.client.name)
        self.ui.lineEditIcon.setText(self.client.icon_name)
        self.ui.lineEditLabel.setText(self.client.label)
        self.ui.checkBoxSaveStop.setChecked(self.client.check_last_save)
        self.ui.toolButtonIcon.setIcon(
            ray.getAppIcon(self.client.icon_name, self))
        self.ui.lineEditIgnoredExtensions.setText(
            self.client.ignored_extensions)

    def changeIconwithText(self, text):
        self.ui.toolButtonIcon.setIcon(ray.getAppIcon(text, self))

    def editExecutable(self):
        dialog = EditExecutableDialog(self, self.client)
        dialog.exec()
        if dialog.result():
            self.ui.labelExecutable.setText(dialog.getExecutable())
            self.ui.labelArguments.setText(dialog.getArguments())

    def saveChanges(self):
        self.client.executable_path = self.ui.labelExecutable.text()
        self.client.arguments = self.ui.labelArguments.text()
        self.client.label = self.ui.lineEditLabel.text()
        self.client.icon_name = self.ui.lineEditIcon.text()
        self.client.check_last_save = self.ui.checkBoxSaveStop.isChecked()
        self.client.ignored_extensions = self.ui.lineEditIgnoredExtensions.text()
        self.client.sendPropertiesToDaemon()
        # better for user to wait a little before close the window
        QTimer.singleShot(150, self.accept)


class ClientTrashDialog(ChildDialog):
    def __init__(self, parent, client_data):
        ChildDialog.__init__(self, parent)
        self.ui = ui_client_trash.Ui_Dialog()
        self.ui.setupUi(self)

        self.client_data = client_data

        self.ui.labelExecutable.setText(self.client_data.executable_path)
        self.ui.labelId.setText(self.client_data.client_id)
        self.ui.labelClientName.setText(self.client_data.name)
        self.ui.labelClientIcon.setText(self.client_data.icon)
        self.ui.labelClientLabel.setText(self.client_data.label)
        self.ui.checkBoxSaveStop.setChecked(self.client_data.check_last_save)
        self.ui.toolButtonIcon.setIcon(QIcon.fromTheme(self.client_data.icon))

        self.ui.pushButtonRemove.clicked.connect(self.removeClient)

    def serverStatusChanged(self, server_status):
        if server_status in (ray.ServerStatus.CLOSE,
                             ray.ServerStatus.OFF,
                             ray.ServerStatus.OUT_SAVE,
                             ray.ServerStatus.OUT_SNAPSHOT,
                             ray.ServerStatus.WAIT_USER):
            self.reject()

    def removeClient(self):
        self.toDaemon(
            '/ray/trash/remove_definitely',
            self.client_data.client_id)
        self.reject()


class AbortSessionDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_abort_session.Ui_AbortSession()
        self.ui.setupUi(self)

        self.ui.pushButtonAbort.clicked.connect(self.accept)
        self.ui.pushButtonCancel.clicked.connect(self.reject)
        self.ui.pushButtonCancel.setFocus(Qt.OtherFocusReason)

        self.serverStatusChanged(self._session.server_status)

    def serverStatusChanged(self, server_status):
        self.ui.pushButtonAbort.setEnabled(
            not bool(
                server_status in (
                    ray.ServerStatus.CLOSE,
                    ray.ServerStatus.OFF,
                    ray.ServerStatus.COPY)))
        if server_status == ray.ServerStatus.OFF:
            self.reject()


class AbortServerCopyDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_abort_copy.Ui_Dialog()
        self.ui.setupUi(self)

        self._signaler.server_progress.connect(self.setProgress)

        self.serverStatusChanged(self._session.server_status)

    def serverStatusChanged(self, server_status):
        if server_status not in (
                ray.ServerStatus.PRECOPY,
                ray.ServerStatus.COPY):
            self.reject()

    def setProgress(self, progress):
        self.ui.progressBar.setValue(progress * 100)


class AbortClientCopyDialog(ChildDialog):
    def __init__(self, parent, client_id):
        ChildDialog.__init__(self, parent)
        self.ui = ui_abort_copy.Ui_Dialog()
        self.ui.setupUi(self)

        self.client_id = client_id

        self._signaler.client_progress.connect(self.setProgress)

    def setProgress(self, client_id, progress):
        if client_id != self.client_id:
            return

        self.ui.progressBar.setValue(progress * 100)

    def serverStatusChanged(self, server_status):
        if not self.server_copying:
            self.reject()


class OpenNsmSessionInfoDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_nsm_open_info.Ui_Dialog()
        self.ui.setupUi(self)

        self.ui.checkBox.stateChanged.connect(self.showThis)

    def showThis(self, state):
        RS.settings.setValue('OpenNsmSessionInfo', not bool(state))


class QuitAppDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_quit_app.Ui_DialogQuitApp()
        self.ui.setupUi(self)
        self.ui.pushButtonCancel.setFocus(Qt.OtherFocusReason)
        self.ui.pushButtonSaveQuit.clicked.connect(self.closeSession)
        self.ui.pushButtonQuitNoSave.clicked.connect(self.abortSession)

        original_text = self.ui.labelExecutable.text()
        self.ui.labelExecutable.setText(
            original_text %
            ('<strong>%s</strong>' %
             self._session.name))

        self.serverStatusChanged(self._session.server_status)

    def serverStatusChanged(self, server_status):
        if server_status == ray.ServerStatus.OFF:
            self.accept()
            return

        self.ui.pushButtonSaveQuit.setEnabled(
            bool(server_status == ray.ServerStatus.READY))
        self.ui.pushButtonQuitNoSave.setEnabled(
            bool(server_status != ray.ServerStatus.CLOSE))

    def closeSession(self):
        self.toDaemon('/ray/session/close')

    def abortSession(self):
        self.toDaemon('/ray/session/abort')


class AboutRaySessionDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_about_raysession.Ui_DialogAboutRaysession()
        self.ui.setupUi(self)
        all_text = self.ui.labelRayAndVersion.text()
        self.ui.labelRayAndVersion.setText(all_text % ray.VERSION)


class NewExecutableDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_new_executable.Ui_DialogNewExecutable()
        self.ui.setupUi(self)
        
        self.ui.groupBoxAdvanced.setVisible(False)
        self.resize(0, 0)
        self.ui.labelPrefixMode.setToolTip(
            self.ui.comboBoxPrefixMode.toolTip())
        self.ui.labelClientId.setToolTip(self.ui.lineEditClientId.toolTip())
        
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.ui.lineEdit.setFocus(Qt.OtherFocusReason)
        self.ui.lineEdit.textChanged.connect(self.textChanged)

        self.ui.checkBoxProxy.stateChanged.connect(self.proxyStateChanged)
        
        self.ui.lineEditPrefix.setEnabled(False)
        self.ui.toolButtonAdvanced.clicked.connect(self.showAdvanced)
        
        self.ui.comboBoxPrefixMode.addItem(
            _translate('new_executable', 'Custom'))
        self.ui.comboBoxPrefixMode.addItem(
            _translate('new_executable', 'Client Name'))
        self.ui.comboBoxPrefixMode.addItem(
            _translate('new_executable', 'Session Name'))
        self.ui.comboBoxPrefixMode.setCurrentIndex(2)
        
        self.ui.comboBoxPrefixMode.currentIndexChanged.connect(
            self.prefixModeChanged)
        
        self._signaler.new_executable.connect(self.addExecutableToCompleter)
        self.toDaemon('/ray/server/list_path')

        self.exec_list = []

        self.completer = QCompleter(self.exec_list)
        self.ui.lineEdit.setCompleter(self.completer)

        self.ui.lineEdit.returnPressed.connect(self.closeNow)
        
        self.serverStatusChanged(self._session.server_status)
        
        self.text_will_accept = False

    def showAdvanced(self):
        self.ui.groupBoxAdvanced.setVisible(True)
        self.ui.toolButtonAdvanced.setVisible(False)
    
    def prefixModeChanged(self, index):
        self.ui.lineEditPrefix.setEnabled(bool(index == 0))
    
    def addExecutableToCompleter(self, executable_list):
        self.exec_list += executable_list
        self.exec_list.sort()

        del self.completer
        self.completer = QCompleter(self.exec_list)
        self.ui.lineEdit.setCompleter(self.completer)

    def getExecutableSelected(self):
        return self.ui.lineEdit.text()

    def runViaProxy(self):
        return bool(self.ui.checkBoxProxy.isChecked())
    
    def getSelection(self):
        return (self.ui.lineEdit.text(),
                self.ui.checkBoxProxy.isChecked(),
                self.ui.comboBoxPrefixMode.currentIndex(),
                self.ui.lineEditPrefix.text(),
                self.ui.lineEditClientId.text())
    
    def proxyStateChanged(self, state):
        self.ui.buttonBox.button(
            QDialogButtonBox.Ok).setEnabled(
            state or self.text_will_accept)

    def textChanged(self, text):
        self.text_will_accept = bool(text)
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(
            self.ui.checkBoxProxy.isChecked() or self.text_will_accept)

    def closeNow(self):
        if (self.ui.lineEdit.text() in self.exec_list
                or self.ui.checkBoxProxy.isChecked()):
            self.accept()

    def serverStatusChanged(self, server_status):
        if server_status in (ray.ServerStatus.OUT_SAVE,
                             ray.ServerStatus.OUT_SNAPSHOT,
                             ray.ServerStatus.WAIT_USER,
                             ray.ServerStatus.CLOSE,
                             ray.ServerStatus.OFF):
            self.reject()


class StopClientDialog(ChildDialog):
    def __init__(self, parent, client_id):
        ChildDialog.__init__(self, parent)
        self.ui = ui_stop_client.Ui_Dialog()
        self.ui.setupUi(self)

        self.client_id = client_id
        self.wait_for_save = False

        self.client = self._session.getClient(client_id)

        if self.client:
            text = self.ui.label.text() % self.client.prettierName()

            if not self.client.has_dirty:
                minutes = int((time.time() - self.client.last_save) / 60)
                text = _translate(
                    'client_stop',
                    "<strong>%s</strong> seems to has not been saved for %i minute(s).<br />Do you really want to stop it ?") % (self.client.prettierName(),
                                                                                                                                 minutes)

            self.ui.label.setText(text)
            
            self.client.status_changed.connect(self.serverUpdatesClientStatus)

        self.ui.pushButtonSaveStop.clicked.connect(self.saveAndStop)
        self.ui.checkBox.stateChanged.connect(self.checkBoxClicked)

        

    def saveAndStop(self):
        self.wait_for_save = True
        self.toDaemon('/ray/client/save', self.client_id)

    def checkBoxClicked(self, state):
        self.client.check_last_save = not bool(state)
        self.client.sendPropertiesToDaemon()

    def serverUpdatesClientStatus(self, status):
        if status in (ray.ClientStatus.STOPPED, ray.ClientStatus.REMOVED):
            self.reject()
            return

        if status == ray.ClientStatus.READY and self.wait_for_save:
            self.wait_for_save = False
            self.accept()


class StopClientNoSaveDialog(ChildDialog):
    def __init__(self, parent, client_id):
        ChildDialog.__init__(self, parent)
        self.ui = ui_stop_client_no_save.Ui_Dialog()
        self.ui.setupUi(self)
        
        self.client_id = client_id
        self.client = self._session.getClient(client_id)

        if self.client:
            text = self.ui.label.text() % self.client.prettierName()
            self.ui.label.setText(text)
            self.client.status_changed.connect(self.serverUpdatesClientStatus)
        
        self.ui.checkBox.stateChanged.connect(self.checkBoxClicked)
        self.ui.pushButtonCancel.setFocus(True)
        
    def serverUpdatesClientStatus(self, status):
        if status in (ray.ClientStatus.STOPPED, ray.ClientStatus.REMOVED):
            self.reject()
            return
        
    def checkBoxClicked(self, state):
        self.client.check_last_save = not bool(state)
        self.client.sendPropertiesToDaemon()


class SnapShotProgressDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_snapshot_progress.Ui_Dialog()
        self.ui.setupUi(self)
        self._signaler.server_progress.connect(self.serverProgress)
        
    def serverStatusChanged(self, server_status):
        self.close()
        
    def serverProgress(self, value):
        self.ui.progressBar.setValue(value * 100)
        

class DaemonUrlWindow(ChildDialog):
    def __init__(self, parent, err_code, ex_url):
        ChildDialog.__init__(self, parent)
        self.ui = ui_daemon_url.Ui_Dialog()
        self.ui.setupUi(self)

        self.ui.lineEdit.textChanged.connect(self.allowUrl)

        error_text = ''
        if err_code == ErrDaemon.NO_ANNOUNCE:
            error_text = _translate(
                "url_window",
                "<p>daemon at<br><strong>%s</strong><br>didn't announce !<br></p>") % ex_url
        elif err_code == ErrDaemon.NOT_OFF:
            error_text = _translate(
                "url_window",
                "<p>daemon at<br><strong>%s</strong><br>has a loaded self._session.<br>It can't be used for slave session</p>") % ex_url
        elif err_code == ErrDaemon.WRONG_ROOT:
            error_text = _translate(
                "url_window",
                "<p>daemon at<br><strong>%s</strong><br>uses an other session root folder !<.p>") % ex_url
        elif err_code == ErrDaemon.FORBIDDEN_ROOT:
            error_text = _translate(
                "url_window",
                "<p>daemon at<br><strong>%s</strong><br>uses a forbidden session root folder !<.p>") % ex_url
        elif err_code == ErrDaemon.WRONG_VERSION:
            error_text = _translate(
                "url_window",
                "<p>daemon at<br><strong>%s</strong><br>uses another %s version.<.p>") % (ex_url, ray.APP_TITLE)
        else:
            error_text = _translate("url window", "<p align=\"left\">To run a network session,<br>open a terminal on another computer of this network.<br>Launch ray-daemon on port 1234 (for example)<br>by typing the command :</p><p align=\"left\"><code>ray-daemon -p 1234</code></p><p align=\"left\">Then paste below the first url<br>that ray-daemon gives you at startup.</p><p></p>")

        self.ui.labelError.setText(error_text)
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)

        self.tried_urls = ray.getListInSettings(RS.settings, 'network/tried_urls')
        last_tried_url = RS.settings.value('network/last_tried_url', '', type=str)

        self.completer = QCompleter(self.tried_urls)
        self.ui.lineEdit.setCompleter(self.completer)

        if ex_url:
            self.ui.lineEdit.setText(ex_url)
        elif last_tried_url:
            self.ui.lineEdit.setText(last_tried_url)

    def getDaemonUrl(self):
        return self.ui.lineEdit.text()

    def allowUrl(self, text):
        if not text:
            self.ui.lineEdit.completer().complete()
            self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
            return

        if not text.startswith('osc.udp://'):
            self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
            return

        try:
            addr = ray.getLibloAddress(text)
            self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(True)
        except BaseException:
            self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)

    def getUrl(self):
        return self.ui.lineEdit.text()


class WaitingCloseUserDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_waiting_close_user.Ui_Dialog()
        self.ui.setupUi(self)
        
        if isDarkTheme(self):
            self.ui.labelSaveIcon.setPixmap(
                QPixmap(':scalable/breeze-dark/document-nosave.svg'))
        
        self.ui.pushButtonOk.setFocus(True)
        self.ui.pushButtonUndo.clicked.connect(self.undoClose)
        self.ui.pushButtonSkip.clicked.connect(self.skip)
        self.ui.checkBox.setChecked(
            bool(RS.settings.value(
                'hide_wait_close_user_dialog', False, type=bool)))
        
        self.ui.checkBox.clicked.connect(self.checkBoxClicked)
        
    def serverStatusChanged(self, server_status):
        if server_status != ray.ServerStatus.WAIT_USER:
            self.accept()
    
    def undoClose(self):
        self.toDaemon('/ray/session/cancel_close')
    
    def skip(self):
        self.toDaemon('/ray/session/skip_wait_user')
        
    def checkBoxClicked(self, state):
        RS.settings.setValue('hide_wait_close_user_dialog',
                             bool(state), type=bool)

class DonationsDialog(ChildDialog):
    def __init__(self, parent, display_no_again):
        ChildDialog.__init__(self, parent)
        self.ui = ui_donations.Ui_Dialog()
        self.ui.setupUi(self)
        
        print('display_no_again', display_no_again)
        self.ui.checkBox.setVisible(display_no_again)
        self.ui.checkBox.clicked.connect(self.checkBoxClicked)
        
    def checkBoxClicked(self, state):
        RS.settings.setValue('hide_donations', state)
        

class ErrorDialog(ChildDialog):
    def __init__(self, parent, message):
        ChildDialog.__init__(self, parent)
        self.ui = ui_error_dialog.Ui_Dialog()
        self.ui.setupUi(self)

        self.ui.label.setText(message)




import os
import subprocess
from PyQt5.QtCore import QProcess, QTimer
from osc_server_thread import OscServerThread
from server_sender import ServerSender
import ray

class CopyFile(object):
    slots = ['orig_path',
             'dest_path',
             'state',
             'size']

class FileCopier(ServerSender):
    def __init__(self, session):
        ServerSender.__init__(self)
        self.session        = session
        self.client_id      = ''
        self.next_function  = None
        self.abort_function = None
        self.next_args      = []
        self.copy_files     = []
        self.copy_size      = 0
        self.aborted        = False
        self.is_active      = False
        
        self.process = QProcess()
        self.process.finished.connect(self.processFinished)
        if ray.QT_VERSION >= (5, 6):
            self.process.errorOccurred.connect(self.errorOccurred)
        
        self.timer = QTimer()
        self.timer.setInterval(250)
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(self.checkProgressSize)
        
    def informCopytoGui(self, copy_state):
        server = OscServerThread.getInstance()
        if not server:
            return
        
        server.informCopytoGui(copy_state)
    
    def getFileSize(self, filepath):
        if not os.path.exists(filepath):
            return 0
        
        try:
            du_full = subprocess.check_output(
                ['nice', '-n', '15', 'du', '-sb', filepath]).decode()
        except:
            du_full = ""
            
        if not du_full:
            return 0
        
        du_str = du_full.split('\t')[0]
        if not du_str.isdigit():
            return 0
        
        return int(du_str)
        
    def checkProgressSize(self):
        current_size = 0
        self.timer.stop()
        
        for copy_file in self.copy_files:
            if copy_file.state == 2:
                current_size += copy_file.size
            elif copy_file.state == 1:
                current_size += self.getFileSize(copy_file.dest_path)
                break

        if current_size and self.copy_size:
            progress = float(current_size/self.copy_size)
            
            if self.client_id:
                self.sendGui('/ray/gui/client/progress', self.client_id, progress)
            else:
                self.sendGui('/ray/gui/server/progress', progress)
                
            self.session.oscReply('/ray/net_daemon/duplicate_state', progress)
        
        self.timer.start()
    
    def processFinished(self, exit_code, exit_status):
        self.timer.stop()
        
        for copy_file in self.copy_files:
            if copy_file.state == 1:
                copy_file.state = 2
                break
        
        if self.aborted:
            ##remove all created files
            for copy_file in self.copy_files:
                if copy_file.state > 0:
                    file_to_remove = copy_file.dest_path
                    
                    if os.path.exists(file_to_remove):
                        try:
                            subprocess.run(['rm', '-R', file_to_remove]) 
                        except:
                            pass
                        
            self.is_active = False
            self.informCopytoGui(False)
            self.abort_function(*self.next_args)
            return
        
        #run next_function if copy is terminated
        for copy_file in self.copy_files:
            if copy_file.state != 2:
                break
        else:
            self.is_active = False
            self.informCopytoGui(False)
            
            if self.next_function:
                self.next_function(*self.next_args)
                
            return
        
        self.nextProcess()
        
    def errorOccurred(self):
        #todo make something else
        self.processFinished(0, 0)
        
    def nextProcess(self):
        self.is_active = True
        
        for copy_file in self.copy_files:
            if copy_file.state == 0:
                copy_file.state = 1
                self.process.start('nice' , 
                                   ['-n', '+15', 'cp', '-R', 
                                    copy_file.orig_path, copy_file.dest_path])
                break
            
        self.timer.start()
    
    def start(self, src_list, dest_dir, next_function,
              abort_function, next_args=[]):
        self.abort_function = abort_function
        self.next_function  = next_function
        self.next_args      = next_args
        
        self.aborted = False
        self.copy_size = 0
        self.copy_files.clear()
        
        dest_path_exists = bool(os.path.exists(dest_dir))
        if dest_path_exists:
            if not os.path.isdir(dest_dir):
                #TODO send error, but it should not append
                self.abort_function(*self.next_args)
                return
        
        if type(src_list) == str:
            src_dir = src_list
            src_list = []
            
            if not os.path.isdir(src_dir):
                self.abort_function(*self.next_args)
                return
            
            try:
                tmp_list = os.listdir(src_dir)
            except:
                self.abort_function(*self.next_args)
                return
                
            for path in tmp_list:
                if path == '.ray-snapshots':
                    continue
                
                full_path = "%s/%s" % (src_dir, path)
                src_list.append(full_path)
            
            if not dest_path_exists:
                os.makedirs(dest_dir)
        
        for orig_path in src_list:
            copy_file = CopyFile()
            copy_file.state     = 0
            copy_file.orig_path = orig_path
            copy_file.size      = self.getFileSize(orig_path)
            
            self.copy_size+=copy_file.size
            
            if dest_path_exists:
                copy_file.dest_path = "%s/%s" % (dest_dir,
                                                 os.path.basename(orig_path))
            else:
                #WARNING works only with one file !!!
                copy_file.dest_path = dest_dir
            
            self.copy_files.append(copy_file)
        
        
        if self.copy_files:
            self.informCopytoGui(True)
            self.nextProcess()
        else:
            self.next_function(*self.next_args)
        
    def startClientCopy(self, client_id, src_list, dest_dir, next_function,
                        abort_function, next_args=[]):
        self.client_id = client_id
        self.start(src_list, dest_dir, next_function,
                   abort_function, next_args)
        
    def startSessionCopy(self, src_dir, dest_dir, next_function,
                         abort_function, next_args=[]):
        self.client_id = ''
        self.start(src_dir, dest_dir, next_function,
                   abort_function, next_args)
        
    def abort(self, abort_function=None, next_args=[]):
        if abort_function:
            self.abort_function = abort_function
            self.next_args = next_args
        
        self.timer.stop()
        if self.process.state() == QProcess.Running:
            self.aborted = True
            self.process.terminate()
            
    def isActive(self, client_id=''):
        if client_id and client_id != self.client_id:
            return False
        
        return self.is_active
 




import ray
import nsm_client
from gui_tools import CommandLineArgs, _translate
from gui_server_thread import GUIServerThread

class NSMChild:
    def __init__(self, session):
        self._session = session
        self.nsm_signaler = nsm_client.NSMSignaler()
        self.nsm_signaler.server_sends_open.connect(self.open)
        self.nsm_signaler.server_sends_save.connect(self.save)
        self.nsm_signaler.show_optional_gui.connect(self.showOptionalGui)
        self.nsm_signaler.hide_optional_gui.connect(self.hideOptionalGui)
        
        self.wait_for_open = False
        self.wait_for_save = False
        self.project_path  = ''
        
        serverNSM = nsm_client.NSMThread('raysession_child',
                                         self.nsm_signaler,
                                         CommandLineArgs.NSM_URL,
                                         CommandLineArgs.debug)        
        serverNSM.start()
        
        self._session._signaler.daemon_announce_ok.connect(
            self.announceToParent)
        self._session._signaler.server_status_changed.connect(
            self.serverStatusChanged)
        
    def announceToParent(self):
        serverNSM = nsm_client.NSMThread.instance()
        
        if serverNSM:
            serverNSM.announce(_translate('child_session', 'Child Session'),
                               ':switch:optional-gui:', 'raysession')
    
    def serverStatusChanged(self, server_status):
        if server_status == ray.ServerStatus.READY:
            serverNSM = nsm_client.NSMThread.instance()
            if not serverNSM:
                return
            
            if self.wait_for_open:
                serverNSM.openReply()
                self.wait_for_open = False
            
            elif self.wait_for_save:
                serverNSM.saveReply()
                self.wait_for_save = False
    
    def open(self, project_path, session_name, jack_client_name):
        self.wait_for_open = True
        self.project_path  = project_path
        
        server = GUIServerThread.instance()
        if server:
            server.openSession(project_path)
    
    def save(self):
        if self._session._main_win:
            self._session._main_win.saveWindowSettings()
        
        self.wait_for_save = True
        
        server = GUIServerThread.instance()
        if server:
            server.saveSession()
        
    def showOptionalGui(self):
        if self._session._main_win:
            self._session._main_win.show()
            
        serverNSM = nsm_client.NSMThread.instance()
        if serverNSM:
            serverNSM.sendGuiState(True)
        
    def hideOptionalGui(self):
        if self._session._main_win:
            self._session._main_win.hide()
            
        serverNSM = nsm_client.NSMThread.instance()
        
        if serverNSM:
            serverNSM.sendGuiState(False)
        
        
class NSMChildOutside(NSMChild):
    def __init__(self, session):
        NSMChild.__init__(self, session)
        self.wait_for_close = False
        
        self.session_name = ''
        self.template_name = ''
        
    def announceToParent(self):
        serverNSM = nsm_client.NSMThread.instance()
        
        if serverNSM:
            serverNSM.announce(_translate('network_session',
                                           'Network Session'),
                                ':switch:optional-gui:ray-network:',
                                'ray-network')
        
        
            serverNSM.sendToDaemon(
                '/nsm/client/network_properties',
                self._session._daemon_manager.url,
                self._session._daemon_manager.session_root)
    
    def save(self):
        serverNSM = nsm_client.NSMThread.instance()
        
        if serverNSM:
            serverNSM.sendToDaemon(
                '/nsm/client/network_properties',
                self._session._daemon_manager.url,
                self._session._daemon_manager.session_root)
        
        NSMChild.save(self)
        
    def open(self, project_path, session_name, jack_client_name):
        self.wait_for_open = True
        
        #Here project_path is used for template if needed
        self.template_name = project_path
        self.session_name  = session_name
        
        server = GUIServerThread.instance()
        if server:
            server.openSession(self.session_name, self.template_name)
        
    def closeSession(self):
        self.wait_for_close = True
        
        server = GUIServerThread.instance()
        if server:
            server.closeSession()



import argparse
import os
import sys
from PyQt5.QtCore import QCoreApplication, QStandardPaths, QSettings

import ray

settings = QSettings()
#non_active_clients = []

def dirname(*args):
    return os.path.dirname(*args)

def getAppConfigPath():
    return "%s/%s" % (
            QStandardPaths.writableLocation(QStandardPaths.ConfigLocation),
            QCoreApplication.organizationName())

def getCodeRoot():
    return dirname(dirname(dirname(os.path.realpath(__file__))))

def initDaemonTools():
    #global non_active_clients
    #del non_active_clients
    
    if CommandLineArgs.config_dir:
        settings = QSettings(CommandLineArgs.config_dir)
    else:
        settings = QSettings()
        
    RS.setSettings(settings)
    
    RS.setNonActiveClients(ray.getListInSettings(settings, 
                                                 'daemon/non_active_list'))
    RS.setFavorites(ray.getListInSettings(settings, 'daemon/favorites'))
    TemplateRoots.initConfig()

def getGitDefaultUnAndIgnored(executable):
    ignored = []
    unignored = []
        
    if executable in ('luppp', 'sooperlooper', 'sooperlooper_nsm'):
        unignored.append('.wav')
            
    elif executable == 'samplv1_jack':
        unignored = ['.wav', '.flac', '.ogg', '.mp3']
        
    return (ignored, unignored)
    
class RS:
    settings = QSettings()
    non_active_clients = []
    favorites = []
    
    @classmethod
    def setSettings(cls, settings):
        del cls.settings
        cls.settings = settings
        
    @classmethod
    def setNonActiveClients(cls, nalist):
        del cls.non_active_clients
        cls.non_active_clients = nalist
        
    @classmethod
    def setFavorites(cls, favorites):
        cls.favorites = favorites

class TemplateRoots:
    net_session_name = ".ray-net-session-templates"
    factory_sessions = "%s/session_templates" % getCodeRoot()
    factory_clients  = "%s/client_templates"  % getCodeRoot()
    
    @classmethod
    def initConfig(cls):
        if CommandLineArgs.config_dir:
            app_config_path = CommandLineArgs.config_dir
        else:
            app_config_path = getAppConfigPath()
            
        cls.user_sessions = "%s/session_templates" % app_config_path
        cls.user_clients  = "%s/client_templates"  % app_config_path


class Terminal:
    _last_client_name = ''
    
    @classmethod
    def message(cls, string):
        if cls._last_client_name and cls._last_client_name != 'daemon':
            sys.stderr.write('\n')
        
        sys.stderr.write('[\033[90mray-daemon\033[0m]\033[92m%s\033[0m\n'
                            % string)
        
        cls._last_client_name = 'daemon'
        
    @classmethod
    def snapshoterMessage(cls, byte_string, command=''):
        snapshoter_str = "snapshoter:.%s" % command
        
        if cls._last_client_name != snapshoter_str:
            sys.stderr.write('\n[\033[90mray-daemon-git%s\033[0m]\n'
                             % command)
        sys.stderr.buffer.write(byte_string)
        
        cls._last_client_name = snapshoter_str

    @classmethod
    def clientMessage(cls, byte_string, client_name, client_id):
        client_str = "%s.%s" % (client_name, client_id)
        
        if not CommandLineArgs.debug_only:
            if cls._last_client_name != client_str:
                sys.stderr.write('\n[\033[90m%s-%s\033[0m]\n'
                                    % (client_name, client_id))
            sys.stderr.buffer.write(byte_string)
            
        cls._last_client_name = client_str
        
    @classmethod
    def warning(cls, string):
        sys.stderr.write('[\033[90mray-daemon\033[0m]%s\033[0m\n' % string)
        cls._last_client_name = 'daemon'


class CommandLineArgs(argparse.Namespace):
    session_root = ''
    osc_port     = 0
    findfreeport = True
    gui_url      = None
    config_dir   = ''
    debug        = False
    debug_only   = False
    session      = ''
    
    @classmethod
    def eatAttributes(cls, parsed_args):
        for attr_name in dir(parsed_args):
            if not attr_name.startswith('_'):
                setattr(cls, attr_name, getattr(parsed_args, attr_name))
                
        if cls.debug_only:
            cls.debug = True
            
        if cls.osc_port == 0:
            cls.osc_port = 16187
            cls.findfreeport = True
            
        if cls.config_dir and not os.access(cls.config_dir, os.W_OK):
            sys.stderr.write(
                '%s is not a writable config dir, try another one\n'
                    % cls.config_dir)
            sys.exit(1)

class ArgParser(argparse.ArgumentParser):
    def __init__(self):
        argparse.ArgumentParser.__init__(self)
        _translate = QCoreApplication.translate
        
        default_root = "%s/%s" % (os.getenv('HOME'), 
                                  _translate('daemon', 
                                             'Ray Network Sessions'))
        
        self.add_argument('--session-root', '-r', type=str, default=default_root,
                          help='set root folder for sessions')
        self.add_argument('--session', '-s', type=str, default='',
                          help='session to load at startup')
        self.add_argument('--osc-port', '-p', 
                          type=int, default=0, 
                          help='select OSC port for the daemon')
        self.add_argument('--findfreeport', action='store_true', 
                          help='find another port if port is not free')
        self.add_argument('--gui-url', type=ray.getLibloAddress, 
                          help=argparse.SUPPRESS)
        self.add_argument('--config-dir', '-c', type=str, default='', 
                          help='use a custom config dir')
        self.add_argument('--debug','-d',  action='store_true', 
                          help='see all OSC messages')
        self.add_argument('--debug-only', '-do', action='store_true', 
                          help='debug without client messages')
        
        self.add_argument('-v', '--version', action='version',
                          version=ray.VERSION)
        
        parsed_args = argparse.ArgumentParser.parse_args(self)
        CommandLineArgs.eatAttributes(parsed_args)
        
        



from PyQt5.QtCore import Qt, QTimer, QSize
from PyQt5.QtWidgets import QDialogButtonBox, QListWidgetItem, QFrame, QMenu, QAction
from PyQt5.QtGui import QIcon, QPalette

import ray

from gui_tools import RS, _translate
from child_dialogs import ChildDialog

import ui_add_application
import ui_template_slot
import ui_remove_template

class TemplateSlot(QFrame):
    def __init__(self, list_widget, item, session, icon, name, factory):
        QFrame.__init__(self)
        self.ui = ui_template_slot.Ui_Frame()
        self.ui.setupUi(self)
        
        self.list_widget = list_widget
        self.item = item
        self._session = session
        self.name = name
        self.icon_name = icon
        self.factory = factory
        
        self.ui.toolButtonIcon.setIcon(ray.getAppIcon(icon, self))
        self.ui.label.setText(name)
        self.ui.toolButtonUser.setVisible(not factory)
        
        self.user_menu = QMenu()
        act_remove_template = QAction(QIcon.fromTheme('edit-delete-remove'),
                                      _translate('menu', 'remove'),
                                      self.user_menu)
        act_remove_template.triggered.connect(self.removeTemplate)
        self.user_menu.addAction(act_remove_template)
        self.ui.toolButtonUser.setMenu(self.user_menu)
        
        self.is_favorite = False
        
        self.favicon_not = QIcon(':scalable/breeze/draw-star.svg')
        
        if (self.palette().brush(2, QPalette.WindowText).color().lightness()
                > 128):
            self.ui.toolButtonUser.setIcon(
                QIcon(':scalable/breeze-dark/im-user.svg'))
            self.favicon_not = QIcon(':scalable/breeze-dark/draw-star.svg')
        
        for favorite in self._session.favorite_list:
            if favorite.name == name and favorite.factory == factory:
                self.is_favorite = True
                break
        
        if self.is_favorite:
            self.ui.toolButtonFavorite.setIcon(
                QIcon(':scalable/breeze/star-yellow.svg'))
        else:
            self.ui.toolButtonFavorite.setIcon(self.favicon_not)
        
        self.ui.toolButtonFavorite.clicked.connect(self.favoriteClicked)
    
    def favoriteClicked(self):
        self.is_favorite = not self.is_favorite
        
        if self.is_favorite:
            self.ui.toolButtonFavorite.setIcon(
                QIcon(':scalable/breeze/star-yellow.svg'))
            self._session.addFavorite(self.name, self.icon_name, self.factory)
        else:
            self.ui.toolButtonFavorite.setIcon(self.favicon_not)
            self._session.removeFavorite(self.name, self.factory)
        
        self.list_widget.setCurrentItem(self.item)
    
    def removeTemplate(self):
        add_app_dialog = self.list_widget.parent()
        add_app_dialog.removeTemplate(self.name, self.factory)
        
    def mouseDoubleClickEvent(self, event):
        self.list_widget.parent().accept()
        
     
class TemplateItem(QListWidgetItem):
    def __init__(self, parent, session, icon, name, factory):
        QListWidgetItem.__init__(self, parent, QListWidgetItem.UserType + 1)
        self.f_widget = TemplateSlot(parent, self, session,
                                     icon, name, factory)
        self.setData(Qt.UserRole, name)
        parent.setItemWidget(self, self.f_widget)
        self.setSizeHint(QSize(100, 28))
        
        self.f_factory = factory
        
    def __lt__(self, other):
        self_name = self.data(Qt.UserRole)
        other_name = other.data(Qt.UserRole)
        
        if other_name == None:
            return False
        
        if self_name == other_name:
            # make the user template on top
            return not self.f_factory
            
        return bool(self.data(Qt.UserRole).lower() < other.data(Qt.UserRole).lower())


class RemoveTemplateDialog(ChildDialog):
    def __init__(self, parent, template_name):
        ChildDialog.__init__(self, parent)
        self.ui = ui_remove_template.Ui_Dialog()
        self.ui.setupUi(self)
        
        self.ui.label.setText(
            _translate('add_app_dialog', 
                       '<p>Are you sure to want to remove<br>the template "%s" and all its files ?</p>')
                            % template_name)
            
        self.ui.pushButtonCancel.setFocus()

class AddApplicationDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_add_application.Ui_DialogAddApplication()
        self.ui.setupUi(self)
        
        self._session = parent._session

        self.ui.checkBoxFactory.setChecked(RS.settings.value(
            'AddApplication/factory_box', True, type=bool))
        self.ui.checkBoxUser.setChecked(RS.settings.value(
            'AddApplication/user_box', True, type=bool))

        self.ui.checkBoxFactory.stateChanged.connect(self.factoryBoxChanged)
        self.ui.checkBoxUser.stateChanged.connect(self.userBoxChanged)

        self.ui.templateList.currentItemChanged.connect(
            self.currentItemChanged)
        self.ui.templateList.setFocus(Qt.OtherFocusReason)
        self.ui.filterBar.textEdited.connect(self.updateFilteredList)
        self.ui.filterBar.updownpressed.connect(self.updownPressed)
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)

        self._signaler.user_client_template_found.connect(
            self.addUserTemplates)
        self._signaler.factory_client_template_found.connect(
            self.addFactoryTemplates)
        self.toDaemon('/ray/server/list_user_client_templates')
        self.toDaemon('/ray/server/list_factory_client_templates')

        self.user_template_list = []
        self.factory_template_list = []

        self.server_will_accept = False
        self.has_selection = False

        self.serverStatusChanged(self._session.server_status)

    def factoryBoxChanged(self, state):
        if not state:
            self.ui.checkBoxUser.setChecked(True)

        self.updateFilteredList()

    def userBoxChanged(self, state):
        if not state:
            self.ui.checkBoxFactory.setChecked(True)

        self.updateFilteredList()

    def serverStatusChanged(self, server_status):
        self.server_will_accept = bool(
            server_status not in (
                ray.ServerStatus.OFF,
                ray.ServerStatus.CLOSE) and not self.server_copying)
        self.preventOk()

    def addUserTemplates(self, template_list):
        for template in template_list:
            template_name = template
            icon_name = ''

            if '/' in template:
                template_name, slash, icon_name = template.partition('/')
                
            if template_name in self.user_template_list:
                continue

            self.user_template_list.append(template_name)
            
            list_widget = TemplateItem(self.ui.templateList,
                                       self._session,
                                       icon_name,
                                       template_name, 
                                       False)

            self.ui.templateList.addItem(list_widget)
            
            self.ui.templateList.sortItems()

        self.updateFilteredList()

    def addFactoryTemplates(self, template_list):
        for template in template_list:
            template_name = template
            icon_name = ''

            if '/' in template:
                template_name, slash, icon_name = template.partition('/')
                
            if template_name in self.factory_template_list:
                continue

            self.factory_template_list.append(template_name)
            
            list_widget = TemplateItem(self.ui.templateList,
                                       self._session,
                                       icon_name,
                                       template_name, 
                                       True)

            self.ui.templateList.addItem(list_widget)
            self.ui.templateList.sortItems()

        self.updateFilteredList()

    def updateFilteredList(self, filt=''):
        filter_text = self.ui.filterBar.displayText()

        # show all items
        for i in range(self.ui.templateList.count()):
            self.ui.templateList.item(i).setHidden(False)

        # hide all non matching items
        for i in range(self.ui.templateList.count()):
            item = self.ui.templateList.item(i)
            template_name = item.data(Qt.UserRole)
            
            if not filter_text.lower() in template_name.lower():
                item.setHidden(True)
            
            if item.f_factory and not self.ui.checkBoxFactory.isChecked():
                item.setHidden(True)
            
            if not item.f_factory and not self.ui.checkBoxUser.isChecked():
                item.setHidden(True)

        # if selected item not in list, then select the first visible
        if (not self.ui.templateList.currentItem()
                or self.ui.templateList.currentItem().isHidden()):
            for i in range(self.ui.templateList.count()):
                if not self.ui.templateList.item(i).isHidden():
                    self.ui.templateList.setCurrentRow(i)
                    break

        if (not self.ui.templateList.currentItem()
                or self.ui.templateList.currentItem().isHidden()):
            self.ui.filterBar.setStyleSheet(
                "QLineEdit { background-color: red}")
            self.ui.templateList.setCurrentItem(None)
        else:
            self.ui.filterBar.setStyleSheet("")
            self.ui.templateList.scrollTo(self.ui.templateList.currentIndex())

    def updownPressed(self, key):
        row = self.ui.templateList.currentRow()
        if key == Qt.Key_Up:
            if row == 0:
                return
            row -= 1
            while self.ui.templateList.item(row).isHidden():
                if row == 0:
                    return
                row -= 1
        elif key == Qt.Key_Down:
            if row == self.ui.templateList.count() - 1:
                return
            row += 1
            while self.ui.templateList.item(row).isHidden():
                if row == self.ui.templateList.count() - 1:
                    return
                row += 1
        self.ui.templateList.setCurrentRow(row)

    def currentItemChanged(self, item, previous_item):
        self.has_selection = bool(item)
        self.preventOk()

    def preventOk(self):
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(
            bool(self.server_will_accept and self.has_selection))

    def getSelectedTemplate(self):
        item = self.ui.templateList.currentItem()
        if item:
            return (item.data(Qt.UserRole), item.f_factory)
        
    def removeTemplate(self, template_name, factory):
        dialog = RemoveTemplateDialog(self, template_name)
        dialog.exec()
        if not dialog.result():
            return 
        
        self.toDaemon('/ray/server/remove_client_template', template_name)
        
        for i in range(self.ui.templateList.count()):
            item = self.ui.templateList.item(i)
            
            if not item.f_factory and template_name == item.data(Qt.UserRole):
                item.setHidden(True)
                self.ui.templateList.removeItemWidget(item)
                break

    def saveCheckBoxes(self):
        RS.settings.setValue(
            'AddApplication/factory_box',
            self.ui.checkBoxFactory.isChecked())
        RS.settings.setValue(
            'AddApplication/user_box',
            self.ui.checkBoxUser.isChecked())
        RS.settings.sync()



#!/usr/bin/python3 -u

import argparse
import os
import sys
import time
import signal
import shutil
import subprocess
from liblo import ServerThread, Address, make_method, Message
from PyQt5.QtCore import (pyqtSignal, QObject, QTimer, QProcess, QSettings,
                          QLocale, QTranslator, QFile)
from PyQt5.QtWidgets import (QApplication, QDialog, QFileDialog, QMessageBox,
                             QMainWindow)
from PyQt5.QtXml import QDomDocument

import ray
import nsm_client
import ui_proxy_gui
import ui_proxy_copy

ERR_OK = 0
ERR_NO_PROXY_FILE = -1
ERR_NOT_ENOUGHT_LINES = -2
ERR_NO_EXECUTABLE = -3
ERR_WRONG_ARGUMENTS = -4
ERR_WRONG_SAVE_SIGNAL = -5
ERR_WRONG_STOP_SIGNAL = -6



def signalHandler(sig, frame):
    if sig in (signal.SIGINT, signal.SIGTERM):
        if proxy.isRunning():
            proxy.waitForStop()
            proxy.stopProcess()
        else:
            #sys.exit()
            app.quit()
        


def ifDebug(string):
    if debug:
        sys.stderr.write(string + '\n')


class ProxyCopyDialog(QDialog):
    def __init__(self):
        QDialog.__init__(self)
        self.ui = ui_proxy_copy.Ui_Dialog()
        self.ui.setupUi(self)

        self.rename_file = False
        self.ui.pushButtonCopyRename.clicked.connect(self.setRenameFile)

    def setRenameFile(self):
        self.rename_file = True
        self.accept()

    def setFile(self, path):
        self.ui.labelFileNotInFolder.setText(
            _translate(
                'Dialog', '%s is not in proxy directory') %
            ('<strong>' + os.path.basename(path) + '</strong>'))


class ProxyDialog(QMainWindow):
    def __init__(self, executable=''):
        QMainWindow.__init__(self)
        self.ui = ui_proxy_gui.Ui_MainWindow()
        self.ui.setupUi(self)
        
        self.server = server
        self.proxy = proxy

        self.config_file = ''
        self.args_edited = False
        self.fields_allow_start = False
        self.process_is_running = False

        
        self.ui.toolButtonBrowse.clicked.connect(self.browseFile)

        self.ui.lineEditExecutable.textEdited.connect(
            self.lineEditExecutableEdited)
        self.ui.lineEditArguments.textChanged.connect(
            self.lineEditArgumentsChanged)
        self.ui.lineEditConfigFile.textChanged.connect(
            self.lineEditConfigFileChanged)
        
        self.ui.comboSaveSig.addItem(_translate('proxy', 'None'), 0)
        self.ui.comboSaveSig.addItem('SIGUSR1', int(signal.SIGUSR1))
        self.ui.comboSaveSig.addItem('SIGUSR2', int(signal.SIGUSR2))
        self.ui.comboSaveSig.addItem('SIGINT',  int(signal.SIGINT))
        self.ui.comboSaveSig.activated.connect(self.comboSaveSigChanged)
        self.ui.comboSaveSig.setCurrentIndex(0)
        
        self.ui.comboNoSave.addItem(
            _translate('proxy', "0 - Ignore missing save"))
        self.ui.comboNoSave.addItem(
            _translate('proxy', "1 - Transmit missing save"))
        self.ui.comboNoSave.addItem(
            _translate('proxy', "2 - Accept windows close"))
        self.ui.comboNoSave.setCurrentIndex(2)
        self.ui.comboNoSave.activated.connect(self.comboNoSaveChanged)
        self.ui.labelNoSaveLevel.setToolTip(self.ui.comboNoSave.toolTip())
        
        self.ui.comboStopSig.addItem('SIGTERM', int(signal.SIGTERM))
        self.ui.comboStopSig.addItem('SIGINT', int(signal.SIGINT))
        self.ui.comboStopSig.addItem('SIGHUP', int(signal.SIGHUP))
        self.ui.comboStopSig.activated.connect(self.comboStopSigChanged)
        self.ui.comboStopSig.setCurrentIndex(0)
        
        self.ui.comboSaveSig.currentTextChanged.connect(self.allowSaveTest)
        self.ui.toolButtonTestSave.clicked.connect(self.testSave)
        self.ui.toolButtonTestSave.setEnabled(False)

        self.ui.pushButtonStart.clicked.connect(self.startProcess)
        self.ui.pushButtonStop.clicked.connect(self.stopProcess)
        self.ui.pushButtonStop.setEnabled(False)

        self.ui.lineEditExecutable.setText(executable)
        self.lineEditExecutableEdited(executable)

        self.ui.labelError.setText('')

        proxy.process.started.connect(self.proxyStarted)
        proxy.process.finished.connect(self.proxyFinished)
        if ray.QT_VERSION >= (5, 6):
            proxy.process.errorOccurred.connect(self.proxyErrorInProcess)

    def checkAllowStart(self):
        self.fields_allow_start = True
        if not self.ui.lineEditExecutable.text():
            self.fields_allow_start = False
        
        if ray.shellLineToArgs(self.ui.lineEditArguments.text()) is None:
            self.fields_allow_start = False

        self.ui.pushButtonStart.setEnabled(
            bool(not self.process_is_running and self.fields_allow_start))

    def updateValuesFromProxyFile(self):
        self.ui.lineEditExecutable.setText(proxy.executable)
        self.ui.lineEditConfigFile.setText(proxy.config_file)
        self.ui.lineEditArguments.setText(proxy.arguments_line)
                
        save_index = self.ui.comboSaveSig.findData(proxy.save_signal)
        self.ui.comboSaveSig.setCurrentIndex(save_index)

        self.ui.comboNoSave.setCurrentIndex(proxy.no_save_level)
        self.ui.comboNoSave.setEnabled(not bool(proxy.save_signal))
        
        stop_index = self.ui.comboStopSig.findData(proxy.stop_signal)
        self.ui.comboStopSig.setCurrentIndex(stop_index)

        self.ui.checkBoxWaitWindow.setChecked(proxy.wait_window)

        self.checkAllowStart()

    def browseFile(self):
        config_file, ok = QFileDialog.getOpenFileName(
            self, _translate('Dialog', 'Select File to use as CONFIG_FILE'))
        if not ok:
            return

        if not config_file.startswith(os.getcwd() + '/'):
            qfile = QFile(config_file)
            if qfile.size() < 20971520:  # if file < 20Mb
                copy_dialog = ProxyCopyDialog()
                copy_dialog.setFile(config_file)
                copy_dialog.exec()

                if copy_dialog.result():
                    if copy_dialog.rename_file:
                        base, pt, extension = os.path.basename(
                            config_file).rpartition('.')

                        config_file = "%s.%s" % (proxy.session_name, extension)
                        if not base:
                            config_file = proxy.session_name
                    else:
                        config_file = os.path.basename(config_file)

                    qfile.copy(config_file)

        self.config_file = os.path.relpath(config_file)
        if (proxy.session_name
            and (self.config_file == proxy.session_name
                 or self.config_file.startswith("%s." % proxy.session_name))):
            self.config_file = self.config_file.replace(proxy.session_name,
                                                        "$RAY_SESSION_NAME")
        self.ui.lineEditConfigFile.setText(self.config_file)

    def lineEditExecutableEdited(self, text):
        self.checkAllowStart()

    def lineEditArgumentsChanged(self, text):
        self.checkAllowStart()
        if ray.shellLineToArgs(text) is not None:
            self.ui.lineEditArguments.setStyleSheet('')
        else:
            self.ui.lineEditArguments.setStyleSheet(
                'QLineEdit{background: red}')
            self.ui.pushButtonStart.setEnabled(False)

    def lineEditConfigFileChanged(self, text):
        if text and not self.ui.lineEditArguments.text():
            self.ui.lineEditArguments.setText('"$CONFIG_FILE"')
        elif (not text 
              and self.ui.lineEditArguments.text() == '"$CONFIG_FILE"'):
            self.ui.lineEditArguments.setText('')

    def comboSaveSigChanged(self, index):
        self.ui.comboNoSave.setEnabled(bool(index == 0))
        
        save_signal = 0
        
        if index == 1:
            save_signal = signal.SIGUSR1
        elif index == 2:
            save_signal = signal.SIGUSR2
        elif index == 3:
            save_signal = signal.SIGINT
            
        save_signal = int(save_signal)
        self.proxy.setSaveSignal(save_signal)
    
    def comboStopSigChanged(self, index):
        stop_signal = signal.SIGTERM
        
        if index == 1:
            stop_signal = signal.SIGINT
        elif index == 2:
            stop_signal = signal.SIGTERM
            
        stop_signal = int(stop_signal)
        self.proxy.setStopSignal(stop_signal)
    
    def comboNoSaveChanged(self, index):
        self.proxy.no_save_level = index
        self.proxy.sendNoSaveLevel()
    
    def allowSaveTest(self, text=None):
        if text is None:
            text = self.ui.comboSaveSig.currentText()

        self.ui.toolButtonTestSave.setEnabled(
            bool(self.process_is_running and text != 'None'))

    def testSave(self):
        save_signal = self.ui.comboSaveSig.currentData()
        proxy.saveProcess(save_signal)

    def saveProxy(self):
        executable = self.ui.lineEditExecutable.text()
        config_file = self.ui.lineEditConfigFile.text()
        arguments_line = self.ui.lineEditArguments.text()
        save_signal = self.ui.comboSaveSig.currentData()
        no_save_level = self.ui.comboNoSave.currentIndex()
        stop_signal = self.ui.comboStopSig.currentData()
        wait_window = self.ui.checkBoxWaitWindow.isChecked()

        proxy.updateAndSave(
            executable,
            config_file,
            arguments_line,
            save_signal,
            no_save_level,
            stop_signal,
            wait_window)

    def startProcess(self):
        self.saveProxy()

        if proxy.is_launchable:
            proxy.startProcess()

    def stopProcess(self):
        proxy.stopProcess(self.ui.comboStopSig.currentData())

    def proxyStarted(self):
        self.process_is_running = True
        self.ui.pushButtonStart.setEnabled(False)
        self.ui.pushButtonStop.setEnabled(True)
        self.allowSaveTest()
        self.ui.labelError.setText('')

    def processTerminateShortly(self, duration):
        self.ui.labelError.setText('Process terminate in %f ms')

    def proxyFinished(self):
        self.process_is_running = False
        self.ui.pushButtonStart.setEnabled(self.fields_allow_start)
        self.ui.pushButtonStop.setEnabled(False)
        self.allowSaveTest()
        self.ui.labelError.setText('')

    def proxyErrorInProcess(self):
        self.ui.labelError.setText(
            _translate(
                'Dialog',
                'Executable failed to launch ! It\'s maybe not present on system.'))
        if not self.isVisible():
            self.show()

    def closeEvent(self, event):
        server.sendToDaemon('/nsm/client/gui_is_hidden')
        settings.setValue(
            'ProxyGui%s/geometry' %
            self.proxy.full_client_id,
            self.saveGeometry())
        settings.setValue(
            'ProxyGui%s/WindowState' %
            self.proxy.full_client_id, self.saveState())
        settings.sync()

        if self.fields_allow_start:
            self.saveProxy()

        QMainWindow.closeEvent(self, event)

        # Quit if process is not running yet
        if not proxy.process.state() == QProcess.Running:
            sys.exit(0)

    def showEvent(self, event):
        self.server.sendToDaemon('/nsm/client/gui_is_shown')

        if settings.value('ProxyGui%s/geometry' % self.proxy.full_client_id):
            self.restoreGeometry(
                settings.value(
                    'ProxyGui%s/geometry' %
                    self.proxy.full_client_id))
        if settings.value('ProxyGui%s/WindowState' % self.proxy.full_client_id):
            self.restoreState(
                settings.value(
                    'ProxyGui%s/WindowState' %
                    self.proxy.full_client_id))

        self.updateValuesFromProxyFile()

        QMainWindow.showEvent(self, event)
##########################


class Proxy(QObject):
    def __init__(self, executable=''):
        QObject.__init__(self)
        self.process = QProcess()
        self.process.setProcessChannelMode(QProcess.ForwardedChannels)
        self.process.finished.connect(self.processFinished)

        #self.proxy_file = None
        self.is_launchable = False
        self.project_path = ""
        self.path = ""
        self.session_name = ""
        self.full_client_id = ""

        self.executable = executable
        self.arguments = []
        self.arguments_line = ''
        self.config_file = ""
        self.save_signal = 0
        self.no_save_level = 2
        self.stop_signal = int(signal.SIGTERM)
        self.label = ""
        
        self.wait_window = False

        self._config_file_used = False
        self._wait_for_stop = False
        
        self.timer_save = QTimer()
        self.timer_save.setSingleShot(True)
        self.timer_save.setInterval(300)
        self.timer_save.timeout.connect(self.timerSaveFinished)

        self.timer_open = QTimer()
        self.timer_open.setSingleShot(True)
        self.timer_open.setInterval(500)
        self.timer_open.timeout.connect(self.timerOpenFinished)

        self.is_finishable = False
        self.timer_close = QTimer()
        self.timer_close.setSingleShot(True)
        self.timer_close.setInterval(2500)
        self.timer_close.timeout.connect(self.timerCloseFinished)
        self.timer_close.start()
        self.process_start_time = time.time()

        self.timer_window = QTimer()
        self.timer_window.setInterval(100)
        self.timer_window.timeout.connect(self.checkWindow)
        self.timer_window_n = 0

        signaler.server_sends_open.connect(self.initialize)
        signaler.server_sends_save.connect(self.saveProcess)
        signaler.show_optional_gui.connect(self.showOptionalGui)
        signaler.hide_optional_gui.connect(self.hideOptionalGui)

    def isRunning(self):
        return bool(self.process.state() == QProcess.Running)

    def waitForStop(self):
        self._wait_for_stop = True
    
    def setSaveSignal(self, int_signal):
        self.save_signal = int_signal
        self.sendNoSaveLevel()
        
    def setStopSignal(self, int_signal):
        self.stop_signal = int_signal
    
    def readFile(self):
        self.is_launchable = False
        try:
            file = open(self.path, 'r')
        except BaseException:
            return

        xml = QDomDocument()
        xml.setContent(file.read())

        content = xml.documentElement()

        if content.tagName() != "RAY-PROXY":
            file.close()
            return

        cte = content.toElement()
        file_version = cte.attribute('VERSION')
        self.executable = cte.attribute('executable')
        self.config_file = cte.attribute('config_file')
        self.arguments_line = cte.attribute('arguments')
        save_signal = cte.attribute('save_signal')
        no_save_level = cte.attribute('no_save_level')
        stop_signal = cte.attribute('stop_signal')

        wait_window = cte.attribute('wait_window')

        if wait_window.isdigit():
            self.wait_window = bool(int(wait_window))
        else:
            self.wait_window = False

        file.close()

        if save_signal.isdigit():
            self.save_signal = int(save_signal)
        
        if no_save_level.isdigit():
            self.no_save_level = int(no_save_level)
        else:
            self.no_save_level = 2
        
        versions = [file_version, '0.7.1']
        versions.sort()
        
        if file_version != versions[0]:
            # something was wrong in old version,
            # save signal was saved as stop signal too.
            # so don't read stop signal if this is an old file.
            if stop_signal.isdigit():
                self.stop_signal = int(stop_signal)

        if not self.executable:
            return

        arguments = ray.shellLineToArgs(self.arguments_line)
        if arguments is None:
            return

        self.is_launchable = True

    def saveFile(
            self,
            executable,
            config_file,
            arguments_line,
            save_signal,
            no_save_level,
            stop_signal,
            wait_window):
        try:
            file = open(self.path, 'w')
        except BaseException:
            return

        if not save_signal:
            save_signal = 0

        xml = QDomDocument()
        p = xml.createElement('RAY-PROXY')
        p.setAttribute('VERSION', ray.VERSION)
        p.setAttribute('executable', executable)
        p.setAttribute('arguments', arguments_line)
        p.setAttribute('config_file', config_file)
        p.setAttribute('save_signal', str(int(save_signal)))
        p.setAttribute('no_save_level', str(no_save_level))
        p.setAttribute('stop_signal', str(int(stop_signal)))
        p.setAttribute('wait_window', wait_window)

        xml.appendChild(p)

        contents = "<?xml version='1.0' encoding='UTF-8'?>\n"
        contents += "<!DOCTYPE RAY-PROXY>\n"
        contents += xml.toString()

        file.write(contents)
        file.close()

        self.readFile()
    
    def updateValues(self, executable, config_file, arguments_line,
                     save_signal, no_save_level, stop_signal, wait_window):
        self.executable = executable
        self.config_file = config_file
        self.arguments_line = arguments_line
        self.save_signal = save_signal
        self.no_save_level = no_save_level
        self.stop_signal = stop_signal
        self.wait_window = wait_window
    
    def saveProxyFile(self):
        self.saveFile(
            self.executable,
            self.config_file,
            self.arguments_line,
            self.save_signal,
            self.no_save_level,
            self.stop_signal,
            self.wait_window)
    
    def updateAndSave(self, executable, config_file, arguments_line,
                      save_signal, no_save_level, stop_signal, wait_window):
        self.updateValues(executable, config_file, arguments_line, 
                          save_signal, no_save_level,
                          stop_signal, wait_window)
        self.saveProxyFile()
    
    def processFinished(self, exit_code):
        if self._wait_for_stop:
            app.quit()
            
        if self.is_finishable:
            if not proxy_dialog.isVisible():
                app.quit()
        else:
            duration = time.time() - self.process_start_time
            proxy_dialog.processTerminateShortly(duration)
            # proxy_dialog.show()
    
    def checkWindow(self):
        self.timer_window_n += 1

        if self.timer_window_n > 600:
            # 600 x 50ms = 30s max until ray-proxy 
            # replyOpen to Session Manager
            self.checkWindowEnded()
            return

        try:
            # get all windows and their PID with wmctrl
            wmctrl_all = subprocess.check_output(
                ['wmctrl', '-l', '-p']).decode()
        except BaseException:
            self.checkWindowEnded()
            return

        if not wmctrl_all:
            self.checkWindowEnded()
            return

        all_lines = wmctrl_all.split('\n')
        pids = []

        # get all windows pids
        for line in all_lines:
            if not line:
                continue

            line_sep = line.split(' ')
            non_empt = []
            for el in line_sep:
                if el:
                    non_empt.append(el)

            if len(non_empt) >= 3 and non_empt[2].isdigit():
                pids.append(int(non_empt[2]))
            else:
                # window manager seems to not work correctly with wmctrl, so
                # replyOpen now
                self.checkWindowEnded()
                return

        parent_pid = self.process.pid()

        # check in pids if one comes from this ray-proxy
        for pid in pids:
            if pid < parent_pid:
                continue

            ppid = pid

            while ppid != parent_pid and ppid > 1:
                try:
                    proc_file = open('/proc/%i/status' % ppid, 'r')
                    proc_contents = proc_file.read()
                except BaseException:
                    self.checkWindowEnded()
                    return
                    
                for line in proc_contents.split('\n'):
                    if line.startswith('PPid:'):
                        ppid_str = line.rpartition('\t')[2]
                        if ppid_str.isdigit():
                            ppid = int(ppid_str)
                            break
                else:
                    self.checkWindowEnded()
                    return
                
            if ppid == parent_pid:
                # a window appears with a pid child of this ray-proxy,
                # replyOpen
                QTimer.singleShot(200, self.checkWindowEnded)
                break

    def checkWindowEnded(self):
        self.timer_window.stop()
        server.openReply()

    def initialize(self, project_path, session_name, full_client_id):
        self.project_path = project_path
        self.session_name = session_name
        self.full_client_id = full_client_id

        server.sendGuiState(False)

        if not os.path.exists(project_path):
            os.mkdir(project_path)

        os.chdir(project_path)

        #proxy_dialog.setWindowTitle(self.full_client_id)
        
        self.path = os.path.join(project_path, "ray-proxy.xml")
        self.readFile()

        proxy_dialog.updateValuesFromProxyFile()

        if not self.is_launchable:
            server.openReply()
            proxy_dialog.show()
            return

        self.startProcess()
    
    def sendNoSaveLevel(self):
        if ':no-save-level:' in server.getServerCapabilities():
            nsl = self.no_save_level
            if self.save_signal:
                nsl = 0
            server.sendToDaemon('/nsm/client/no_save_level', nsl)
    
    def startProcess(self):
        os.environ['NSM_CLIENT_ID'] = self.full_client_id
        os.environ['RAY_SESSION_NAME'] = self.session_name

        # enable environment vars in config_file
        config_file = os.path.expandvars(self.config_file)
        os.environ['CONFIG_FILE'] = config_file

        # Useful for launching NSM compatible clients with specifics arguments
        os.unsetenv('NSM_URL')

        arguments_line = os.path.expandvars(self.arguments_line)
        arguments = ray.shellLineToArgs(arguments_line)
        
        self._config_file_used = bool(config_file
                                      and config_file in arguments)
        self.sendNoSaveLevel()
        
        self.process.start(self.executable, arguments)
        self.timer_open.start()

    def saveProcess(self, save_signal=0):
        if not save_signal:
            save_signal = self.save_signal

        if self.isRunning() and save_signal:
            os.kill(self.process.processId(), save_signal)

        self.timer_save.start()

    def stopProcess(self, signal=signal.SIGTERM):
        if signal is None:
            return

        if not self.isRunning():
            return

        os.kill(self.process.processId(), signal)

    def timerSaveFinished(self):
        server.saveReply()

    def timerOpenFinished(self):
        if self.wait_window:
            self.timer_window.start()
        else:
            server.openReply()

        if self.isRunning() and proxy_dialog.isVisible():
            proxy_dialog.close()

    def timerCloseFinished(self):
        self.is_finishable = True

    def stop(self):
        if self.process.state:
            self.process.terminate()

    def showOptionalGui(self):
        proxy_dialog.show()

    def hideOptionalGui(self):
        if not proxy_dialog.isHidden():
            proxy_dialog.close()


class ProxyFile(object):
    def __init__(self, project_path, executable=''):
        

        self.executable = executable
        self.arguments_line = ''
        self.config_file = ''
        self.args_line = ''
        self.save_signal = 0
        self.stop_signal = int(signal.SIGTERM)
        self.wait_window = False

        

    
    

if __name__ == '__main__':
    NSM_URL = os.getenv('NSM_URL')
    if not NSM_URL:
        sys.stderr.write('Could not register as NSM client.\n')
        sys.exit()

    daemon_address = ray.getLibloAddress(NSM_URL)

    parser = argparse.ArgumentParser()
    parser.add_argument('--executable', default='')
    parser.add_argument('--debug',
                        '-d',
                        action='store_true',
                        help='see all OSC messages')
    parser.add_argument('-v', '--version', action='version',
                        version=ray.VERSION)
    parsed_args = parser.parse_args()

    debug = parsed_args.debug
    executable = parsed_args.executable

    app = QApplication(sys.argv)
    app.setApplicationName("RaySession")
    # app.setApplicationVersion(ray.VERSION)
    app.setOrganizationName("RaySession")
    app.setQuitOnLastWindowClosed(False)
    settings = QSettings()
    
    signal.signal(signal.SIGINT, signalHandler)
    signal.signal(signal.SIGTERM, signalHandler)

    # Translation process
    locale = QLocale.system().name()
    appTranslator = QTranslator()
    
    if appTranslator.load(
        "%s/locale/raysession_%s" %
        (os.path.dirname(
            os.path.dirname(
                os.path.dirname(
                    sys.argv[0]))),
         locale)):
        app.installTranslator(appTranslator)
    _translate = app.translate

    timer = QTimer()
    timer.setInterval(200)
    timer.timeout.connect(lambda: None)
    timer.start()

    signaler = nsm_client.NSMSignaler()
    
    server = nsm_client.NSMThread('ray-proxy', signaler,
                                  daemon_address, debug)
    server.start()
    
    proxy = Proxy(executable)
    proxy_dialog = ProxyDialog()

    
    server.announce('Ray Proxy', ':optional-gui:warning-no-save:', 'ray-proxy')

    app.exec()

    settings.sync()
    server.stop()

    del server
    del proxy
    del app



import os
import socket
import sys
from liblo import Address
from PyQt5.QtCore import QObject, QProcess, QTimer
from PyQt5.QtWidgets import QApplication

import ray
from gui_signaler import Signaler
from gui_server_thread import GUIServerThread
from gui_tools import (CommandLineArgs, default_session_root, ErrDaemon,
                       _translate)

class DaemonManager(QObject):
    def __init__(self, session):
        QObject.__init__(self)
        self._session = session
        self._signaler = self._session._signaler

        self.executable = 'ray-daemon'
        self.process = QProcess()

        self.process.finished.connect(self.processFinished)
        if ray.QT_VERSION >= (5, 6):
            self.process.errorOccurred.connect(self.errorInProcess)
        self.process.setProcessChannelMode(QProcess.ForwardedChannels)

        self.announce_timer = QTimer()
        self.announce_timer.setInterval(2000)
        self.announce_timer.setSingleShot(True)
        self.announce_timer.timeout.connect(self.announceTimerOut)

        self.stopped_yet = False
        self.is_local = True
        self.launched_before = False
        self.address = None
        self.port = None
        self.url = ''

        self.is_announced = False
        self.is_nsm_locked = False

        self._signaler.daemon_announce.connect(self.receiveAnnounce)
        self._signaler.daemon_url_changed.connect(self.changeUrl)

    def finishInit(self):
        self._main_win = self._session._main_win

    def processFinished(self, exit_code, exit_status):
        if self.stopped_yet:
            QApplication.quit()
            return

        if not self._main_win.isHidden():
            self._main_win.daemonCrash()
            return

    def errorInProcess(self, error):
        self._main_win.daemonCrash()

    def changeUrl(self, new_url):
        try:
            self.setOscAddress(ray.getLibloAddress(new_url))
        except BaseException:
            return

        self.callDaemon()

    def callDaemon(self):
        if not self.address:
            # I don't know really why, but it works only with a timer
            QTimer.singleShot(5, self.showDaemonUrlWindow)
            return

        self.announce_timer.start()

        server = GUIServerThread.instance()
        if not server:
            sys.stderr.write(
                'GUI can not call daemon, GUI OSC server is missing.\n')
            return

        server.announce()

    def showDaemonUrlWindow(self):
        self._signaler.daemon_url_request.emit(ErrDaemon.NO_ERROR, self.url)

    def announceTimerOut(self):
        if self.launched_before:
            self._signaler.daemon_url_request.emit(
                ErrDaemon.NO_ANNOUNCE, self.url)
        else:
            sys.stderr.write(
                _translate(
                    'error',
                    "No announce from ray-daemon. RaySession can't works. Sorry.\n"))
            QApplication.quit()

    def receiveAnnounce(
            self,
            src_addr,
            version,
            server_status,
            options,
            session_root,
            is_net_free):
        self.announce_timer.stop()

        if version.split('.')[:1] != ray.VERSION.split('.')[:1]:
            # works only if the two firsts digits are the same (ex: 0.6)
            self._signaler.daemon_url_request.emit(
                ErrDaemon.WRONG_VERSION, self.url)
            self.disannounce(src_addr)
            return

        if CommandLineArgs.net_session_root and session_root != CommandLineArgs.net_session_root:
            self._signaler.daemon_url_request.emit(
                ErrDaemon.WRONG_ROOT, self.url)
            self.disannounce(src_addr)
            return

        if not is_net_free:
            self._signaler.daemon_url_request.emit(
                ErrDaemon.FORBIDDEN_ROOT, self.url)
            self.disannounce(src_addr)
            return

        if CommandLineArgs.out_daemon and server_status != ray.ServerStatus.OFF:
            self._signaler.daemon_url_request.emit(ErrDaemon.NOT_OFF, self.url)
            self.disannounce(src_addr)
            return

        self.is_announced = True
        self.address = src_addr
        self.port = src_addr.port
        self.url = src_addr.url
        self.session_root = session_root

        self.is_nsm_locked = options & ray.Option.NSM_LOCKED
        save_all_from_saved_client = options & ray.Option.SAVE_FROM_CLIENT
        bookmark_session_folder = options & ray.Option.BOOKMARK_SESSION

        if self.is_nsm_locked:
            #self._signaler.daemon_nsm_locked.emit(True)
            self._session._main_win.setNsmLocked(True)
        elif CommandLineArgs.under_nsm:
            server = GUIServerThread.instance()
            server.toDaemon('/ray/server/set_nsm_locked')

        self._signaler.daemon_announce_ok.emit()
        self._session.setDaemonOptions(options)

    def disannounce(self, address=None):
        if not address:
            address = self.address

        if address:
            server = GUIServerThread.instance()
            server.disannounce(address)

        self.port = None
        self.url = ''
        del self.address
        self.address = None
        self.is_announced = False

    def setExternal(self):
        self.launched_before = True

    def setOscAddress(self, address):
        self.address = address
        self.launched_before = True
        self.port = self.address.port
        self.url = self.address.url

        self.is_local = bool(self.address.hostname == socket.gethostname())

    def setOscAddressViaUrl(self, url):
        self.setOscAddress(ray.getLibloAddress(url))

    def processIsRunning(self):
        return bool(self.process.state() == 2)

    def start(self):
        if self.launched_before:
            self.callDaemon()
            return

        server = GUIServerThread.instance()
        if not server:
            sys.stderr.write(
                "impossible for GUI to launch daemon. server missing.\n")

        # start process
        arguments = ['--gui-url', str(server.url),
                     '--osc-port', str(self.port),
                     '--session-root', CommandLineArgs.session_root]
        
        if CommandLineArgs.session:
            arguments.append('--session')
            arguments.append(CommandLineArgs.session)

        if CommandLineArgs.debug_only:
            arguments.append('--debug-only')
        elif CommandLineArgs.debug:
            arguments.append('--debug')

        if CommandLineArgs.config_dir:
            arguments.append('--config-dir')
            arguments.append(CommandLineArgs.config_dir)

        self.process.start('ray-daemon', arguments)
        #self.process.start('konsole', ['-e', 'ray-daemon'] + arguments)

    def stop(self):
        if self.launched_before:
            self.disannounce()
            QApplication.quit()
            return

        if self.processIsRunning():
            if not self.stopped_yet:
                self.process.terminate()
                self.stopped_yet = True
                QTimer.singleShot(5000, self.notEndedAfterWait)

    def notEndedAfterWait(self):
        sys.stderr.write('ray-daemon is still running, sorry !\n')
        QApplication.quit()

    def setNewOscAddress(self):
        if not (self.address or self.port):
            self.port = ray.getFreeOscPort()
            self.address = Address(self.port)

    def setNsmLocked(self):
        self.is_nsm_locked = True

    def isAnnounced(self):
        return self.is_announced

    def setDisannounced(self):
        server = GUIServerThread.instance()
        server.disannounce()

        self.port = None
        self.url = ''
        del self.address
        self.address = None
        self.is_announced = False

    def getUrl(self):
        if self.address:
            return self.address.url

        return ''



import time
from PyQt5.QtWidgets import (QListWidget, QListWidgetItem, QFrame, QMenu, 
                             QLineEdit, QAction)
from PyQt5.QtGui import QIcon, QPalette, QPixmap, QFontMetrics, QFont, QFontDatabase
from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot, QSize, QFile

import ray
from gui_server_thread import GUIServerThread
from gui_tools import clientStatusString, _translate, isDarkTheme
import child_dialogs
import snapshots_dialog

import ui_client_slot



class ClientSlot(QFrame):
    def __init__(self, list_widget, client):
        QFrame.__init__(self)
        self.ui = ui_client_slot.Ui_ClientSlotWidget()
        self.ui.setupUi(self)

        # needed variables
        self.list_widget = list_widget
        self.client = client
        self._main_win = self.client._session._main_win

        self.is_dirty_able = False
        self.gui_visible = True
        self.ui.toolButtonGUI.setVisible(False)

        # connect buttons to functions
        self.ui.toolButtonGUI.toggleGui.connect(self.toggleGui)
        self.ui.startButton.clicked.connect(self.startClient)
        self.ui.stopButton.clicked.connect(self.stopClient)
        self.ui.killButton.clicked.connect(self.killClient)
        self.ui.saveButton.clicked.connect(self.saveClient)
        self.ui.closeButton.clicked.connect(self.trashClient)
        self.ui.lineEditClientStatus.statusPressed.connect(self.abortCopy)
        # self.ui.ClientName.name_changed.connect(self.updateLabel)
        
        # prevent "stopped" status displayed at client switch
        #self.ui.lineEditClientStatus,.setTextNow(
                          #clientStatusString(self.client.status))

        self.icon_on = QIcon()
        self.icon_off = QIcon()

        self.updateClientData()

        self.ui.actionSaveAsApplicationTemplate.triggered.connect(
            self.saveAsApplicationTemplate)
        self.ui.actionProperties.triggered.connect(
            self.client.showPropertiesDialog)
        self.ui.actionReturnToAPreviousState.triggered.connect(
            self.openSnapshotsDialog)

        self.menu = QMenu(self)

        self.menu.addAction(self.ui.actionSaveAsApplicationTemplate)
        self.menu.addAction(self.ui.actionReturnToAPreviousState)
        self.menu.addAction(self.ui.actionProperties)
        
        self.ui.actionReturnToAPreviousState.setVisible(
            self._main_win.has_git)

        self.ui.iconButton.setMenu(self.menu)

        self.saveIcon = QIcon()
        self.saveIcon.addPixmap(
            QPixmap(':scalable/breeze/document-save'),
            QIcon.Normal,
            QIcon.Off)
        self.saveIcon.addPixmap(
            QPixmap(':scalable/breeze/disabled/document-save'),
            QIcon.Disabled,
            QIcon.Off)
        self.ui.saveButton.setIcon(self.saveIcon)

        self.savedIcon = QIcon()
        self.savedIcon.addPixmap(
            QPixmap(':scalable/breeze/document-saved'),
            QIcon.Normal,
            QIcon.Off)

        self.unsavedIcon = QIcon()
        self.unsavedIcon.addPixmap(
            QPixmap(':scalable/breeze/document-unsaved'),
            QIcon.Normal,
            QIcon.Off)
        
        self.noSaveIcon = QIcon()
        self.noSaveIcon.addPixmap(
            QPixmap(':scalable/breeze/document-nosave'),
            QIcon.Normal,
            QIcon.Off)

        # choose button colors
        if isDarkTheme(self):
            startIcon = QIcon()
            startIcon.addPixmap(
                QPixmap(':scalable/breeze-dark/media-playback-start'),
                QIcon.Normal,
                QIcon.Off)
            startIcon.addPixmap(
                QPixmap(':scalable/breeze-dark/disabled/media-playback-start'),
                QIcon.Disabled,
                QIcon.Off)
            self.ui.startButton.setIcon(startIcon)

            stopIcon = QIcon()
            stopIcon.addPixmap(
                QPixmap(':scalable/breeze-dark/media-playback-stop'),
                QIcon.Normal,
                QIcon.Off)
            stopIcon.addPixmap(
                QPixmap(':scalable/breeze-dark/disabled/media-playback-stop'),
                QIcon.Disabled,
                QIcon.Off)
            self.ui.stopButton.setIcon(stopIcon)

            self.saveIcon = QIcon()
            self.saveIcon.addPixmap(
                QPixmap(':scalable/breeze-dark/document-save'),
                QIcon.Normal,
                QIcon.Off)
            self.saveIcon.addPixmap(
                QPixmap(':scalable/breeze-dark/disabled/document-save'),
                QIcon.Disabled,
                QIcon.Off)
            self.ui.saveButton.setIcon(self.saveIcon)

            self.savedIcon = QIcon()
            self.savedIcon.addPixmap(
                QPixmap(':scalable/breeze-dark/document-saved'),
                QIcon.Normal,
                QIcon.Off)

            self.unsavedIcon = QIcon()
            self.unsavedIcon.addPixmap(
                QPixmap(':scalable/breeze-dark/document-unsaved'),
                QIcon.Normal,
                QIcon.Off)
            
            self.noSaveIcon = QIcon()
            self.noSaveIcon.addPixmap(
                QPixmap(':scalable/breeze-dark/document-nosave'),
                QIcon.Normal,
                QIcon.Off)

            closeIcon = QIcon()
            closeIcon.addPixmap(
                QPixmap(':scalable/breeze-dark/window-close'),
                QIcon.Normal,
                QIcon.Off)
            closeIcon.addPixmap(
                QPixmap(':scalable/breeze-dark/disabled/window-close'),
                QIcon.Disabled,
                QIcon.Off)
            self.ui.closeButton.setIcon(closeIcon)

        self.ubuntu_font = QFont(
            QFontDatabase.applicationFontFamilies(0)[0], 8)
        self.ubuntu_font_cond = QFont(
            QFontDatabase.applicationFontFamilies(1)[0], 8)
        self.ubuntu_font.setBold(True)
        self.ubuntu_font_cond.setBold(True)

        self.ui.killButton.setVisible(False)

    def clientId(self):
        return self.client.client_id

    def toDaemon(self, *args):
        server = GUIServerThread.instance()
        if server:
            server.toDaemon(*args)

    def startClient(self):
        self.toDaemon('/ray/client/resume', self.clientId())

    def stopClient(self):
        # we need to prevent accidental stop with a window confirmation
        # under conditions
        self._main_win.stopClient(self.clientId())

    def killClient(self):
        self.toDaemon('/ray/client/kill', self.clientId())

    def saveClient(self):
        self.toDaemon('/ray/client/save', self.clientId())

    def trashClient(self):
        self.toDaemon('/ray/client/trash', self.clientId())

    def abortCopy(self):
        self._main_win.abortCopyClient(self.clientId())

    def saveAsApplicationTemplate(self):
        dialog = child_dialogs.SaveTemplateClientDialog(self._main_win)
        dialog.exec()
        if not dialog.result():
            return

        template_name = dialog.getTemplateName()
        self.toDaemon('/ray/client/save_as_template', self.clientId(),
                      template_name)
        
    def openSnapshotsDialog(self):
        dialog = snapshots_dialog.ClientSnapshotsDialog(self._main_win,
                                                        self.client)
        dialog.exec()
        if dialog.result():
            snapshot = dialog.getSelectedSnapshot()
            self.toDaemon('/ray/client/open_snapshot',
                          self.clientId(), snapshot)

    def updateLabel(self, label):
        self._main_win.updateClientLabel(self.clientId(), label)

    def updateClientData(self):
        # set main label
        label = self.client.label if self.client.label else self.client.name
        self.ui.ClientName.setText(label)

        # set tool tip
        self.ui.ClientName.setToolTip(
            'Executable : '
                + self.client.executable_path + '\n'
                + 'Client id : ' + self.clientId())

        # set icon
        self.icon_on = ray.getAppIcon(self.client.icon_name, self)
        self.icon_off = QIcon(self.icon_on.pixmap(32, 32, QIcon.Disabled))

        self.grayIcon(
            bool(
                self.client.status in (
                    ray.ClientStatus.STOPPED,
                    ray.ClientStatus.PRECOPY)))

    def grayIcon(self, gray):
        if gray:
            self.ui.iconButton.setIcon(self.icon_off)
        else:
            self.ui.iconButton.setIcon(self.icon_on)

    def updateStatus(self, status):
        self.ui.lineEditClientStatus.setText(clientStatusString(status))
        
        if status in (
                ray.ClientStatus.LAUNCH,
                ray.ClientStatus.OPEN,
                ray.ClientStatus.SWITCH,
                ray.ClientStatus.NOOP):
            self.ui.startButton.setEnabled(False)
            self.ui.stopButton.setEnabled(True)
            self.ui.saveButton.setEnabled(False)
            self.ui.closeButton.setEnabled(False)
            self.ui.ClientName.setStyleSheet('QLabel {font-weight : bold}')
            self.ui.ClientName.setEnabled(True)
            self.ui.toolButtonGUI.setEnabled(True)
            self.grayIcon(False)

        elif status == ray.ClientStatus.READY:
            self.ui.startButton.setEnabled(False)
            self.ui.stopButton.setEnabled(True)
            self.ui.closeButton.setEnabled(False)
            self.ui.ClientName.setStyleSheet('QLabel {font-weight : bold}')
            self.ui.ClientName.setEnabled(True)
            self.ui.toolButtonGUI.setEnabled(True)
            self.ui.saveButton.setEnabled(True)
            self.grayIcon(False)

        elif status == ray.ClientStatus.STOPPED:
            self.ui.startButton.setEnabled(True)
            self.ui.stopButton.setEnabled(False)
            self.ui.saveButton.setEnabled(False)
            self.ui.closeButton.setEnabled(True)
            self.ui.ClientName.setStyleSheet('QLabel {font-weight : normal}')
            self.ui.ClientName.setEnabled(False)
            self.ui.toolButtonGUI.setEnabled(False)
            self.grayIcon(True)

            self.ui.stopButton.setVisible(True)
            self.ui.killButton.setVisible(False)

            self.ui.saveButton.setIcon(self.saveIcon)

        elif status == ray.ClientStatus.PRECOPY:
            self.ui.startButton.setEnabled(False)
            self.ui.stopButton.setEnabled(False)
            self.ui.saveButton.setEnabled(False)
            self.ui.closeButton.setEnabled(True)
            self.ui.ClientName.setStyleSheet('QLabel {font-weight : normal}')
            self.ui.ClientName.setEnabled(False)
            self.ui.toolButtonGUI.setEnabled(False)
            self.grayIcon(True)

            self.ui.stopButton.setVisible(True)
            self.ui.killButton.setVisible(False)

            self.ui.saveButton.setIcon(self.saveIcon)

        elif status == ray.ClientStatus.COPY:
            self.ui.saveButton.setEnabled(False)

    def allowKill(self):
        self.ui.stopButton.setVisible(False)
        self.ui.killButton.setVisible(True)

    def flashIfOpen(self, boolflash):
        if boolflash:
            self.ui.lineEditClientStatus.setText(
                clientStatusString(ray.ClientStatus.OPEN))
        else:
            self.ui.lineEditClientStatus.setText('')

    def showGuiButton(self):
        self.ui.toolButtonGUI.setVisible(True)
        if self.client.executable_path in ('nsm-proxy', 'ray-proxy'):
            self.ui.toolButtonGUI.setText(_translate('client_slot', 'proxy'))
            self.ui.toolButtonGUI.setToolTip(
                _translate('client_slot', 'Display proxy window'))

    def setGuiState(self, state):
        self.gui_visible = state
        self.ui.toolButtonGUI.setChecked(state)

    def toggleGui(self):
        if not self.gui_visible:
            self.toDaemon('/ray/client/show_optional_gui', self.clientId())
        else:
            self.toDaemon('/ray/client/hide_optional_gui', self.clientId())

    def setDirtyState(self, bool_dirty):
        self.is_dirty_able = True

        if bool_dirty:
            self.ui.saveButton.setIcon(self.unsavedIcon)
        else:
            self.ui.saveButton.setIcon(self.savedIcon)

    def setNoSaveLevel(self, no_save_level):
        if no_save_level:
            self.ui.saveButton.setIcon(self.noSaveIcon)
        else:
            self.ui.saveButton.setIcon(self.saveIcon)
    
    def setProgress(self, progress):
        self.ui.lineEditClientStatus.setProgress(progress)
        
    def setDaemonOptions(self, options):
        has_git = bool(options & ray.Option.HAS_GIT)
        self.ui.actionReturnToAPreviousState.setVisible(has_git)

    def contextMenuEvent(self, event):
        act_selected = self.menu.exec(self.mapToGlobal(event.pos()))
        event.accept()


class ClientItem(QListWidgetItem):
    def __init__(self, parent, client_data):
        QListWidgetItem.__init__(self, parent, QListWidgetItem.UserType + 1)
        self.f_widget = ClientSlot(parent, client_data)
        parent.setItemWidget(self, self.f_widget)
        self.setSizeHint(QSize(100, 45))
        self.sort_number = 0

    def __lt__(self, other):
        result = bool(self.sort_number < other.sort_number)
        return result

    def __gt__(self, other):
        return self.sort_number > other.sort_number

    def setSortNumber(self, sort_number):
        self.sort_number = sort_number

    def getClientId(self):
        return self.f_widget.clientId()


class ListWidgetClients(QListWidget):
    def __init__(self, parent):
        QListWidget.__init__(self, parent)
        self.last_n = 0
        self._session = None
        
    def createClientWidget(self, client_data):
        item = ClientItem(self, client_data)
        item.setSortNumber(self.last_n)
        self.last_n += 1
        return item.f_widget

    def removeClientWidget(self, client_id):
        for i in range(self.count()):
            item = self.item(i)
            if item.getClientId() == client_id:
                widget = item.f_widget
                self.takeItem(i)
                del item
                break

    def reOrderClients(self, client_id_list):
        # when re_order comes from ray-daemon (loading session)
        if len(client_id_list) != self.count():
            return

        for client_id in client_id_list:
            for i in range(self.count()):
                if self.item(i).getClientId() == client_id:
                    break
            else:
                return

        next_item_list = []

        n = 0

        for client_id in client_id_list:
            for i in range(self.count()):
                if self.item(i).getClientId() == client_id:
                    self.item(i).setSortNumber(n)
                    break
            n += 1

        self.sortItems()
    
    def setSession(self, session):
        self._session = session
    
    def toDaemon(self, *args):
        server = GUIServerThread.instance()
        if server:
            server.toDaemon(*args)
    
    @pyqtSlot()
    def launchFavorite(self):
        template_name, factory = self.sender().data()
        self.toDaemon('/ray/session/add_client_template',
                      int(factory),
                      template_name)

    def dropEvent(self, event):
        QListWidget.dropEvent(self, event)

        client_ids_list = []

        for i in range(self.count()):
            item = self.item(i)
            #widget = self.itemWidget(item)
            client_id = item.getClientId()
            client_ids_list.append(client_id)

        server = GUIServerThread.instance()
        if server:
            server.toDaemon('/ray/session/reorder_clients', *client_ids_list)

    def mousePressEvent(self, event):
        if not self.itemAt(event.pos()):
            self.setCurrentRow(-1)
            
        QListWidget.mousePressEvent(self, event)
    
    def contextMenuEvent(self, event):
        if not self.itemAt(event.pos()):
            self.setCurrentRow(-1)
            
            if (self._session
                    and not self._session.server_status in (
                        ray.ServerStatus.OFF,
                        ray.ServerStatus.CLOSE,
                        ray.ServerStatus.OUT_SAVE,
                        ray.ServerStatus.WAIT_USER,
                        ray.ServerStatus.OUT_SNAPSHOT)):
                menu = QMenu()
                fav_menu = QMenu(_translate('menu', 'Favorites'), menu)
                fav_menu.setIcon(QIcon(':scalable/breeze/star-yellow'))
                
                for favorite in self._session.favorite_list:
                    act_app = fav_menu.addAction(
                        ray.getAppIcon(favorite.icon, self), favorite.name)
                    act_app.setData([favorite.name, favorite.factory])
                    act_app.triggered.connect(self.launchFavorite)
            
                menu.addMenu(fav_menu)
                
                menu.addAction(
                    self._session._main_win.ui.actionAddApplication)
                menu.addAction(self._session._main_win.ui.actionAddExecutable)
            
                act_selected = menu.exec(self.mapToGlobal(event.pos()))
            event.accept()
            return
        
        



#!/usr/bin/python3

import sys

file = open(sys.argv[1], 'r')
contents = file.read()
all_lines = contents.split('\n')
output = ""

for line in all_lines:
    while line.startswith('    '):
        line = line.replace('    ', '')
    if line.startswith('@ray_method('):
        raym, par, qmessargs = line.partition('(')
        qmess, qargs = qmessargs.split(',')
        mess = qmess[1:-1]
        if qargs[1:-1] == 'None':
            args = 'strings'
        else:
            args = qargs[2:-2]
        output += "%s %s\n" % (mess, args)

print(output)




import ray
from daemon_manager import DaemonManager
from gui_client import Client, TrashedClient
from gui_signaler import Signaler
from gui_server_thread import GUIServerThread
from gui_tools import initGuiTools, CommandLineArgs, RS
from main_window import MainWindow
from nsm_child import NSMChild, NSMChildOutside


class Session(object):
    def __init__(self):
        self.client_list = []
        self.trashed_clients = []
        self.favorite_list = []
        self.name = ''
        self.path = ''
        self.is_running = False
        self.server_status = ray.ServerStatus.OFF

        self.is_renameable = True
        
        self._signaler = Signaler()
        
        server = GUIServerThread.instance()
        server.start()

        self._daemon_manager = DaemonManager(self)
        if CommandLineArgs.daemon_url:
            self._daemon_manager.setOscAddress(CommandLineArgs.daemon_url)
        elif not CommandLineArgs.out_daemon:
            self._daemon_manager.setNewOscAddress()

        if CommandLineArgs.under_nsm:
            if CommandLineArgs.out_daemon:
                self._nsm_child = NSMChildOutside(self)
                self._daemon_manager.setExternal()
            else:
                self._nsm_child = NSMChild(self)
        
        # build nsm_child if NSM_URL in env
        self._nsm_child = None
        
        if CommandLineArgs.under_nsm:
            if CommandLineArgs.out_daemon:
                self._nsm_child = NSMChildOutside(self)
                self._daemon_manager.setExternal()
            else:
                self._nsm_child = NSMChild(self)
        
        # build and show Main UI
        self._main_win = MainWindow(self)
        
        self._daemon_manager.finishInit()
        server.finishInit(self)
        
        self._main_win.show()
        
        # display donations dialog under conditions
        if not RS.settings.value('hide_donations', False, type=bool):
            coreff_counter = RS.settings.value('coreff_counter', 0, type=int)
            coreff_counter+= 1
            RS.settings.setValue('coreff_counter', coreff_counter)
            
            if coreff_counter % 44 == 29:
                self._main_win.donate(True)

        # The only way I found to not show Messages Dock by default.
        if not RS.settings.value('MainWindow/ShowMessages', False, type=bool):
            self._main_win.hideMessagesDock()

    def quit(self):
        self._main_win.hide()
        del self._main_win

    def setRunning(self, bool):
        self.is_running = bool

    def isRunning(self):
        return bool(self.server_status != ray.ServerStatus.OFF)

    def updateServerStatus(self, server_status):
        self.server_status = server_status

    def setName(self, session_name):
        self.name = session_name
        
    def setPath(self, session_path):
        self.path = session_path
        
    def getShortPath(self):
        if self.path.startswith(CommandLineArgs.session_root):
            return self.path.replace(
                '%s/' % CommandLineArgs.session_root, '', 1)
        
        return self.path

    def getClient(self, client_id):
        for client in self.client_list:
            if client.client_id == client_id:
                return client
        else:
            raise NameError("gui_session does not contains client %s"
                                % client_id)

    def removeAllClients(self):
        self.client_list.clear()
            
    def addFavorite(self, name, icon_name, factory, from_server=False):
        for favorite in self.favorite_list:
            if favorite.name == name and favorite.factory == factory:
                favorite.icon = icon_name
                return 
        
        fav = ray.Favorite(name, icon_name, factory)
        self.favorite_list.append(fav)
        
        self._main_win.updateFavoritesMenu()
        
        if not from_server:
            server = GUIServerThread.instance()
            if server:
                server.toDaemon('/ray/favorites/add', name,
                                icon_name, int(factory))
        
    def removeFavorite(self, name, factory, from_server=False):
        for favorite in self.favorite_list:
            if favorite.name == name and favorite.factory == factory:
                self.favorite_list.remove(favorite)
                break
        
        self._main_win.updateFavoritesMenu()
        
        if not from_server:
            server = GUIServerThread.instance()
            if server:
                server.toDaemon('/ray/favorites/remove', name, int(factory))

    def setDaemonOptions(self, options):
        self._main_win.setDaemonOptions(options)
        for client in self.client_list:
            client.widget.setDaemonOptions(options)

class SignaledSession(Session):
    def __init__(self):
        Session.__init__(self)
        self._signaler.osc_receive.connect(self.oscReceive)
        
        self._daemon_manager.start()
        
    def oscReceive(self, path, args):
        func_path = path
        func_name = func_path.replace('/', '_')
        
        if func_name in self.__dir__():
            function = self.__getattribute__(func_name)
            function(path, args)
    
    def _error(self, path, args):
        err_path, err_code, err_message = args
        self._main_win.errorMessage(err_message)
    
    def _ray_gui_server_nsm_locked(self, path, args):
        nsm_locked = bool(args[0])
        self._main_win.setNsmLocked(nsm_locked)
    
    def _ray_gui_server_message(self, path, args):
        message = args[0]
        self._main_win.printMessage(message)
    
    def _ray_gui_server_options(self, path, args):
        options = args[0]
        self.setDaemonOptions(options)
    
    def _ray_gui_session_name(self, path, args):
        sname, spath = args
        self.setName(sname)
        self.setPath(spath)
        self._main_win.renameSession(sname, spath)
        
    def _ray_gui_session_is_nsm(self, path, args):
        self._main_win.openingNsmSession()
    
    def _ray_gui_session_renameable(self, path, args):
        self.is_renameable = bool(args[0])
        
        bool_set_edit = bool(self.is_renameable
                             and self.server_status == ray.ServerStatus.READY
                             and not CommandLineArgs.out_daemon)
        
        self._main_win.setSessionNameEditable(bool_set_edit)
    
    def _ray_gui_session_sort_clients(self, path, args):
        new_client_list = []
        for client_id in args:
            client = self.getClient(client_id)

            if not client:
                return

            new_client_list.append(client)

        self.client_list.clear()
        self._main_win.reCreateListWidget()

        self.client_list = new_client_list
        for client in self.client_list:
            client.reCreateWidget()
            client.widget.updateStatus(client.status)
    
    def _ray_gui_client_new(self, path, args):
        client = Client(self, ray.ClientData(*args))
        self.client_list.append(client)
    
    def _ray_gui_client_update(self, path, args):
        client_data = ray.ClientData(*args)
        client = self.getClient(client_data.client_id)
        if client:
            client.updateClientProperties(client_data)
    
    def _ray_gui_client_status(self, path, args):
        client_id, status = args
        client = self.getClient(client_id)
        if client:
            client.setStatus(status)
            
            if status == ray.ClientStatus.REMOVED:
                self._main_win.removeClient(client_id)
                client.properties_dialog.close()
                self.client_list.remove(client)
                del client
            
        self._main_win.clientStatusChanged(client_id, status)
    
    def _ray_gui_client_switch(self, path, args):
        old_client_id, new_client_id = args
        
        client = self.getClient(old_client_id)
        if client:
            client.switch(new_client_id)
    
    def _ray_gui_client_progress(self, path, args):
        client_id, progress = args
        
        client = self.getClient(client_id)
        if client:
            client.setProgress(progress)
    
    def _ray_gui_client_dirty(self, path, args):
        client_id, int_dirty = args
        client = self.getClient(client_id)
        if client:
            client.setDirtyState(bool(int_dirty))
    
    def _ray_gui_client_has_optional_gui(self, path, args):
        client_id = args[0]
        client = self.getClient(client_id)
        if client:
            client.setGuiEnabled()
    
    def _ray_gui_client_gui_visible(self, path, args):
        client_id, int_state = args
        client = self.getClient(client_id)
        if client:
            client.setGuiState(bool(int_state))
    
    def _ray_gui_client_still_running(self, path, args):
        client_id = args[0]
        client = self.getClient(client_id)
        if client:
            client.allowKill()
    
    def _ray_gui_client_no_save_level(self, path, args):
        client_id, no_save_level = args
        
        client = self.getClient(client_id)
        if client:
            client.setNoSaveLevel(no_save_level)
    
    def _ray_gui_trash_add(self, path, args):
        client_data = ray.ClientData(*args)
        trash_action = self._main_win.trashAdd(client_data)
        trashed_client = TrashedClient(client_data, trash_action)
        self.trashed_clients.append(trashed_client)
    
    def _ray_gui_trash_remove(self, path, args):
        client_id = args[0]
        
        for trashed_client in self.trashed_clients:
            if trashed_client.data.client_id == client_id:
                break
        else:
            return
        
        self.trashed_clients.remove(trashed_client)
        self._main_win.trashRemove(trashed_client.menu_action)
        
    def _ray_gui_trash_clear(self, path, args):
        self.trashed_clients.clear()
        self._main_win.trashClear()
        
    def _ray_gui_favorites_added(self, path, args):
        name, icon_name, int_factory = args
        self.addFavorite(name, icon_name, bool(int_factory), True)
        
    def _ray_gui_favorites_remove(self, path, args):
        name, int_factory = args
        self.removeFavorite(name, bool(int_factory), True)
        
